'''Chemical Engineering Design Library (ChEDL). Utilities for process modeling.
Copyright (C) 2019, Caleb Bell <Caleb.Andrew.Bell@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''


__all__ = ['standard_entropy', 'S0_basis_converter', 'standard_state_ideal_gas_formation']

from fluids.numerics import quad
from chemicals.reaction import standard_formation_reaction
from thermo.heat_capacity import HeatCapacitySolid, HeatCapacityLiquid, HeatCapacityGas
from chemicals.elements import periodic_table

def standard_entropy(c=None, dS_trans_s=None, dH_trans_s=None, T_trans_s=None,
                     Cp_s_fun=None,
                     Sfusm=None, Hfusm=None, Tm=None, Cp_l_fun=None,
                     Svapm=None, Hvapm=None, Tb=None, Cp_g_fun=None,
                     T_ref=298.15, T_low=0.5, force_gas=True):
    if Tm is None:
        Tm = c.Tm
    if Tb is None:
        Tb = c.Tb
    if Hfusm is None:
        Hfusm = c.Hfusm
    if Hvapm is None:
        Hvapm = c.EnthalpyVaporization(Tb)

    # Misc crystalline transitions
    tot = 0.0
    if dS_trans_s is not None:
        tot += sum(dS_trans_s)
    if dH_trans_s is not None and T_trans_s is not None:
        for dH, T in zip(dH_trans_s, T_trans_s):
            if T < T_ref or force_gas:
                tot += dH/T

    # Solid heat capacity integral
    if Cp_s_fun is not None:
        tot += float(quad(lambda T: Cp_s_fun(T)/T, T_low, Tm)[0])
    else:
        tot += c.HeatCapacitySolid.T_dependent_property_integral_over_T(T_low, Tm)

    # Heat of fusion
    if force_gas or Tm < T_ref:
        if Sfusm is not None:
            tot += Sfusm
        else:
            tot += Hfusm/Tm

    # Liquid heat capacity
    if not force_gas and Tb > T_ref:
        T_liquid_int = T_ref
    else:
        T_liquid_int = Tb
    if force_gas or Tm < T_ref:
        if Cp_l_fun is not None:
            tot += float(quad(lambda T: Cp_l_fun(T)/T, Tm, T_liquid_int)[0])
        else:
            tot += c.HeatCapacityLiquid.T_dependent_property_integral_over_T(Tm, T_liquid_int)

    # Heat of vaporization
    if force_gas or Tb < T_ref:
        if Svapm is not None:
            tot += Svapm
        else:
            tot += Hvapm/Tb

    if force_gas or Tb < T_ref:
        # gas heat capacity
        if Cp_g_fun is not None:
            tot += float(quad(lambda T: Cp_g_fun(T)/T, Tb, T_ref)[0])
        else:
            tot += c.HeatCapacityGas.T_dependent_property_integral_over_T(Tb, T_ref)

    return tot


def S0_basis_converter(c, S0_liq=None, S0_gas=None, T_ref=298.15):
    r'''This function converts a liquid or gas standard entropy to the
    other. This is useful, as thermodynamic packages often work with ideal-
    gas as the reference state and require ideal-gas Gibbs energies of
    formation.

    Parameters
    ----------
    c : Chemical
        Chemical object, [-]
    S0_liq : float, optional
        Liquid absolute entropy of the compound at the reference temperature
        [J/mol/K]
    S0_gas : float, optional
        Gas absolute entropy of the compound at the reference temperature
        [J/mol/K]
    T_ref : float, optional
        The standard state temperature, default 298.15 K; few values are
        tabulated at other temperatures, [-]

    Returns
    -------
    S0_calc : float
        Standard absolute entropy of the compound at the reference temperature
        in the other state to the one provided, [J/mol]

    Notes
    -----
    This function relies in accurate heat capacity curves for both the liquid
    and gas state.

    Examples
    --------
    >>> from thermo.chemical import Chemical
    >>> S0_basis_converter(Chemical('decane'), S0_liq=425.89) # doctest:+SKIP
    544.6792
    >>> S0_basis_converter(Chemical('decane'), S0_gas=545.7) # doctest:+SKIP
    426.9107
    '''
    if S0_liq is None and S0_gas is None:
        raise ValueError("Provide either a liquid or a gas standard absolute entropy")
    if S0_liq is None:
        dS = c.HeatCapacityGas.T_dependent_property_integral_over_T(T_ref, c.Tb)
        dS -= c.EnthalpyVaporization(c.Tb)/c.Tb
        dS += c.HeatCapacityLiquid.T_dependent_property_integral_over_T(c.Tb, T_ref)
        return S0_gas + dS
    else:
        dS = c.HeatCapacityLiquid.T_dependent_property_integral_over_T(T_ref, c.Tb)
        dS += c.EnthalpyVaporization(c.Tb)/c.Tb
        dS += c.HeatCapacityGas.T_dependent_property_integral_over_T(c.Tb, T_ref)
        return S0_liq + dS


standard_state_transitions = {
    'Si': {'Tm': 1687.15, 'Tb': 3504.616, 'Hfus': 50210.0, 'Hvap': 384548.0},
    'Br': {'Tm': 265.900, 'Tb': 332.503, 'Hfus': 10571, 'Hvap': 29563},
    'I': {'Tm': 386.750, 'Tb': 457.666, 'Hfus': 15517, 'Hvap': 41960},
    'B': {'Tm': 2350.000, 'Tb': 4139.449, 'Hfus': 50208, 'Hvap': 480509},
    'Mg': {'Tm': 923.000, 'Tb': 1366.104, 'Hfus': 8477, 'Hvap': 127867},
    'Hg': {'Tm': 234.290, 'Tb': 629.839, 'Hfus': 2295, 'Hvap': 59205},
    'Pb': {'Tm': 600.600, 'Tb': 2019.022, 'Hfus': 4774, 'Hvap': 177582},
    'Li': {'Tm': 453.690, 'Tb': 1620.120, 'Hfus': 3000, 'Hvap': 145843},
    'Na': {'Tm': 370.980, 'Tb': 1170.525, 'Hfus': 2603, 'Hvap': 97022},
    'Al': {'Tm': 933.450, 'Tb': 2790.812, 'Hfus': 10711, 'Hvap': 294002},
    'K': {'Tm': 336.350, 'Tb': 1039.540, 'Hfus': 2334, 'Hvap': 79556},
    'V': {'Tm': 2190.000, 'Tb': 3690.080, 'Hfus': 22845, 'Hvap': 446977},
    'Cr': {'Tm': 2130.000, 'Tb': 2952.078, 'Hfus': 20502, 'Hvap': 339537},
    # 'Mn': {'Tm': 1519.000, 'Tb': 2334.526, 'Hfus': 12058, 'Hvap': 225980,  'solid_T_trans': [980.000, 1361.000, 1412.000], 'solid_H_trans': [2226, 2122, 1879]}, # Double TODO,
    # 'Ti': {'Tm': 1939.000, 'Tb': 3630.956, 'Hfus': 14146, 'Hvap': 409984,  'solid_T_trans': [1166.000], 'solid_H_trans': [4172]}, # TODO
    # 'Ca': {'Tm': 1115.000, 'Tb': 1773.658, 'Hfus': 8540, 'Hvap': 149047,  'solid_T_trans': [716.000], 'solid_H_trans': [930]}, # TODO
    # 'S': {'Tm': 388.360, 'Tb': 882.117, 'Hfus': 1721, 'Hvap': 53326,  'solid_T_trans': [368.300], 'solid_H_trans': [401]}, # TODO
    # 'Be': {'Tm': 1560.000, 'Tb': 2741.437, 'Hfus': 7895, 'Hvap': 291572, 'solid_T_trans': [1527.000], 'solid_H_trans': [6849]}, # TODO
    # 'Fe': {'Tm': 3133.345, 'Hfus': 349585, 'solid_T_trans': [1184.000, 1665.000], 'solid_H_trans': [900, 837]}, # TODO
    
    }

shomate_gas_elements = ('H', 'O', 'N', 'F', 'P', 'Cl', 'Br', 'I', 'Mg', 'B', 'Pb', 'Li', 'Na', 'Al', 'K', 'V', 'Cr')
standard_state_supported_elements = shomate_gas_elements + ('C', 'Si', 'Hg')
standard_state_supported_elements_set = set(standard_state_supported_elements)

element_HeatCapacityGas_dict = {}

def element_HeatCapacityGas_cache(CAS):
    try:
        return element_HeatCapacityGas_dict[CAS]
    except:
        obj = HeatCapacityGas(CASRN=CAS)
        element_HeatCapacityGas_dict[CAS] = obj
        return obj

element_HeatCapacityLiquid_dict = {}

def element_HeatCapacityLiquid_cache(CAS):
    try:
        return element_HeatCapacityLiquid_dict[CAS]
    except:
        obj = HeatCapacityLiquid(CASRN=CAS)
        element_HeatCapacityLiquid_dict[CAS] = obj
        return obj

element_HeatCapacitySolid_dict = {}

def element_HeatCapacitySolid_cache(CAS):
    try:
        return element_HeatCapacitySolid_dict[CAS]
    except:
        obj = HeatCapacitySolid(CASRN=CAS)
        element_HeatCapacitySolid_dict[CAS] = obj
        return obj


def standard_state_ideal_gas_formation(c, T, Hf=None, Sf=None, T_ref=298.15):
    r'''This function calculates the standard state ideal-gas heat of formation
    of a compound at a specified from first principles. The entropy change and
    Gibbs free energy change of formation are also returned. 
    This special condition is usually tabulated in thermodynamic tables, and
    this function is intended to be consistent with the standard conventions.

    The values returned depend on:
    
    * The heat of formation as an ideal gas of the compound at 298.15 K (must be measured experimentally)
    * The ideal gas absolute entropy of the compound at 298.15 K (calculated from heat capacity and phase transitions of the compound and its various phases)
    * The chosen reference states of the elements (convention, but hardcoded)
    * The heat capacity and phase transitions of the constituent elements of the compound in various phases

    Parameters
    ----------
    c : Chemical
        Chemical object, [-]
    T : float
        The temperature to perform the calculation at
    Hf : float, optional
        Standard enthalpy of formation of the compound in the gas phase [J/mol]
    Sf : float, optional
        Standard entropy of formation of the compound in the gas phase [J/mol/K]
    T_ref : float, optional
        The standard state temperature, default 298.15 K; if set to another
        value `Hf` and `Sf` must be provided for this temperature `T_ref`
        or the default values at 298.15 K will be used and this function will
        return nonsense, [-]

    Returns
    -------
    H_standard_state : float
        The standard state ideal gas enthalpy of the compound, [J/mol]
    S_standard_state : float
        The standard state ideal gas entropy of the compound, [J/mol/K]
    G_standard_state : float
        The standard state ideal gas Gibbs energy of the compound, [J/mol]

    Notes
    -----
    Not all elements are supported. The default property methods
    for some phases of the elements are hardcoded in this function for
    accuracy, and may change as data evolves.

    Examples
    --------
    >>> from thermo.chemical import Chemical
    >>> standard_state_ideal_gas_formation(Chemical('water'), T=500.0)
    (-243821.4, -49.63, -219006.)
    '''
    # Whatever the compound is, it is assumed to be in the standard state
    # not that this should not be called on elements
    # Can check against JANAF
    Hf_ref = Hf if Hf is not None else c.Hfgm
    Sf_ref = Sf if Sf is not None else c.Sfgm
    atoms = c.atoms
    return _standard_state_ideal_gas_formation_direct(T=T, Hf_ref=Hf_ref, Sf_ref=Sf_ref, 
            atoms=atoms, gas_Cp=c.HeatCapacityGas, T_ref=T_ref)

_standard_formation_reaction_cache = {}
MAX_STANDARD_FORMATION_CACHE = 250
def _standard_state_ideal_gas_formation_direct(T, Hf_ref, Sf_ref, atoms, gas_Cp, T_ref=298.15, cache=True):
    if cache:
        atoms_key = hash(tuple(atoms.items()))
        if atoms_key in _standard_formation_reaction_cache:
            reactant_coeff, elemental_counts, elemental_composition = _standard_formation_reaction_cache[atoms_key]
        else:
            reactant_coeff, elemental_counts, elemental_composition = standard_formation_reaction(atoms)
            _standard_formation_reaction_cache[atoms_key] = reactant_coeff, elemental_counts, elemental_composition

            if len(_standard_formation_reaction_cache) > MAX_STANDARD_FORMATION_CACHE:
                # Prevent the cache growing too much
                _standard_formation_reaction_cache.pop(next(iter(_standard_formation_reaction_cache)))
    else:
        _standard_formation_reaction_cache[atoms_key] = reactant_coeff, elemental_counts, elemental_composition

    dH_compound = gas_Cp.T_dependent_property_integral(T_ref, T)
    dS_compound = gas_Cp.T_dependent_property_integral_over_T(T_ref, T)
    
    H_calc = reactant_coeff*Hf_ref + reactant_coeff*dH_compound
    S_calc = reactant_coeff*Sf_ref + reactant_coeff*dS_compound
    # if the compound is an element it will need special handling to go from solid liquid to gas if needed

    solid_ele = set(['C'])
    liquid_ele = set([''])
    
    for coeff, ele_data in zip(elemental_counts, elemental_composition):
        ele = list(ele_data.keys())[0]
        element_obj = periodic_table[ele]
#         element = Chemical(element_obj.CAS_standard)
        solid_obj = element_HeatCapacitySolid_cache(element_obj.CAS_standard)
        liquid_obj = element_HeatCapacityLiquid_cache(element_obj.CAS_standard)
        gas_obj = element_HeatCapacityGas_cache(element_obj.CAS_standard)
        if ele not in standard_state_supported_elements_set:
            raise NotImplementedError(f"The element {ele} is not currently supported")

        if ele in shomate_gas_elements:
            gas_obj.method = 'WEBBOOK_SHOMATE'
            if 'WEBBOOK_SHOMATE' in liquid_obj.all_methods:
                liquid_obj.method = 'WEBBOOK_SHOMATE'
            if 'WEBBOOK_SHOMATE' in solid_obj.all_methods:
                solid_obj.method = 'WEBBOOK_SHOMATE'
        elif ele == 'Si':
            solid_obj.method = 'JANAF_FIT'
            liquid_obj.method = 'JANAF_FIT'
            gas_obj.method = 'JANAF_FIT'

        if ele in standard_state_transitions:
            dat = standard_state_transitions[ele]
            Tm = dat['Tm']
            Tb = dat['Tb']
            Hfus = dat['Hfus']
            Hvap = dat['Hvap']
            Tm_solid_int = min(T, Tm)
            T_liquid_int = min(T, Tb)
            dH_ele = solid_obj.T_dependent_property_integral(T_ref, Tm_solid_int)
            dS_ele = solid_obj.T_dependent_property_integral_over_T(T_ref, Tm_solid_int)
            if T > Tm:
                dH_ele += Hfus
                dS_ele += Hfus/Tm
                dH_ele += liquid_obj.T_dependent_property_integral(Tm, T_liquid_int)
                dS_ele += liquid_obj.T_dependent_property_integral_over_T(Tm, T_liquid_int)
            if T > Tb:
                dH_ele += Hvap
                dS_ele += Hvap/Tb
                dH_ele += gas_obj.T_dependent_property_integral(Tb, T)
                dS_ele += gas_obj.T_dependent_property_integral_over_T(Tb, T)
        elif ele == 'P':
            # White phosphorus is the basis here
            T_alpha_beta_P = 195.400
            Htrans_alpha_beta_P = 521.0 # 525.5104 reported in 
            # The thermodynamic properties of elementary phosphorus The heat capacities of two crystalline modifications of red phosphorus, of α and β white phosphorus, and of black phosphorus from 15 to 300 K
            Tm_P = 317.300
            Hfus_P = 659
            Tb_P = 1180.008
            Hvap_P = 63728.0
            # https://janaf.nist.gov/tables/P-001.html
            # ALPHA <--> BETA 195.4 K, BETA <--> LIQUID 317.3 K, LIQUID <--> IDEAL GAS 1180.008 K
            T_solid_int0 = min(T, T_alpha_beta_P)
            T_solid_int1 = min(T, Tm_P)
            T_liquid_int = min(T, Tb_P)
            if T < T_alpha_beta_P:
                dH_ele = solid_obj.T_dependent_property_integral(T_ref, T)
                dS_ele = solid_obj.T_dependent_property_integral_over_T(T_ref, T)

                dH_ele -= Htrans_alpha_beta_P
                dS_ele -= Htrans_alpha_beta_P/T_alpha_beta_P
                # dH_ele -= solid_obj.T_dependent_property_integral(T_alpha_beta_P, T_liquid_int)
                # dS_ele -= solid_obj.T_dependent_property_integral_over_T(Tm_P, T_liquid_int)
            else:
                dH_ele = solid_obj.T_dependent_property_integral(T_ref, T_solid_int1)
                dS_ele = solid_obj.T_dependent_property_integral_over_T(T_ref, T_solid_int1)
                if T > Tm_P:
                    dH_ele += Hfus_P
                    dS_ele += Hfus_P/Tm_P
                    dH_ele += liquid_obj.T_dependent_property_integral(Tm_P, T_liquid_int)
                    dS_ele += liquid_obj.T_dependent_property_integral_over_T(Tm_P, T_liquid_int)
                if T > Tb_P:
                    dH_ele += Hvap_P
                    dS_ele += Hvap_P/Tb_P
                    dH_ele += gas_obj.T_dependent_property_integral(Tb_P, T)
                    dS_ele += gas_obj.T_dependent_property_integral_over_T(Tb_P, T)
        elif ele == 'S':
            # CRystal II to Crystal 1 at 368 K
            # crystal I to liquid at 388 K
            # 432 K liquid-liquid lambda transition
            # 882 K liquid to ideal gas transition
            raise NotImplementedError
        # Need to do all the metals with no fancy phases at once generically
        # https://janaf.nist.gov/tables/Ni-001.html
        # https://janaf.nist.gov/tables/Cu-001.html
        # https://janaf.nist.gov/tables/Zn-001.html

        elif ele in solid_ele:
            dH_ele = solid_obj.T_dependent_property_integral(T_ref, T)
            dS_ele = solid_obj.T_dependent_property_integral_over_T(T_ref, T)
        elif ele in liquid_ele:
            dH_ele = liquid_obj.T_dependent_property_integral(T_ref, T)
            dS_ele = liquid_obj.T_dependent_property_integral_over_T(T_ref, T)
        else:
            dH_ele = gas_obj.T_dependent_property_integral(T_ref, T)
            dS_ele = gas_obj.T_dependent_property_integral_over_T(T_ref, T)
        H_calc -= coeff*dH_ele
        S_calc -= coeff*dS_ele
    G_calc = H_calc - T*S_calc
    
    H_calc, S_calc, G_calc = H_calc/reactant_coeff, S_calc/reactant_coeff, G_calc/reactant_coeff
    
    return H_calc, S_calc, G_calc



'''Chemical Engineering Design Library (ChEDL). Utilities for process modeling.
Copyright (C) 2019, 2020, 2021, 2022 Caleb Bell <Caleb.Andrew.Bell@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

This module contains classes and functions for performing flash calculations.

For reporting bugs, adding feature requests, or submitting pull requests,
please use the `GitHub issue tracker <https://github.com/CalebBell/thermo/>`_.


'''

__all__ = ['Flash']

from math import floor, log10, nan

from chemicals.utils import mixing_simple, property_mass_to_molar, rho_to_Vm
from fluids.constants import R
from fluids.numerics import linspace, logspace
from fluids.numerics import numpy as np

from thermo import phases
from thermo.equilibrium import EquilibriumState
from thermo.flash.flash_utils import (
    LL_boolean_check,
    VL_boolean_check,
    VLL_boolean_check,
    VLL_or_LL_boolean_check,
    VLN_or_LN_boolean_check,
    flash_phase_boundary_one_sided_secant,
    incipient_liquid_bounded_PT_sat,
    incipient_phase_bounded_naive,
    incipient_phase_one_sided_secant,
)
from thermo.phase_identification import identify_sort_phases
from thermo.utils import has_matplotlib

try:
    zeros, ones, ndarray = np.zeros, np.ones, np.ndarray
except:
    pass

spec_to_iter_vars = {
     (True, False, False, True, False, False) : ('T', 'H', 'P'), # Iterating on P is slow, derivatives look OK
     # (True, False, False, True, False, False) : ('T', 'H', 'V'), # Iterating on P is slow, derivatives look OK
     (True, False, False, False, True, False) : ('T', 'S', 'P'),
     (True, False, False, False, False, True) : ('T', 'U', 'V'),

     (False, True, False, True, False, False) : ('P', 'H', 'T'),
     (False, True, False, False, True, False) : ('P', 'S', 'T'),
     (False, True, False, False, False, True) : ('P', 'U', 'T'),

     (False, False, True, True, False, False) : ('V', 'H', 'P'), # TODO: change these ones to iterate on T?
     (False, False, True, False, True, False) : ('V', 'S', 'P'),
     (False, False, True, False, False, True) : ('V', 'U', 'P'),
}

spec_to_iter_vars_backup =  {
    (True, False, False, True, False, False) : ('T', 'H', 'V'),
    (True, False, False, False, True, False) : ('T', 'S', 'V'),
    (True, False, False, False, False, True) : ('T', 'U', 'P'),

    (False, True, False, True, False, False) : ('P', 'H', 'V'),
    (False, True, False, False, True, False) : ('P', 'S', 'V'),
    (False, True, False, False, False, True) : ('P', 'U', 'V'),

    (False, False, True, True, False, False) : ('V', 'H', 'T'),
    (False, False, True, False, True, False) : ('V', 'S', 'T'),
    (False, False, True, False, False, True) : ('V', 'U', 'T'),
}

empty_flash_conv = {'iterations': 0, 'err': 0.0, 'stab_guess_name': None}
one_in_list = [1.0]
empty_list = []

NAIVE_BISECTION_PHASE_MIXING_BOUNDARY = 'NAIVE_BISECTION_PHASE_MIXING_BOUNDARY'
SATURATION_SECANT_PHASE_MIXING_BOUNDARY = 'SATURATION_SECANT_PHASE_MIXING_BOUNDARY'
PT_SECANT_PHASE_MIXING_BOUNDARY = 'SATURATION_SECANT_PHASE_MIXING_BOUNDARY'

SECANT_PHASE_BOUNDARY = 'SECANT_PHASE_BOUNDARY'

CAS_H2O = '7732-18-5'


class Flash:
    r'''Base class for performing flash calculations. All Flash objects need
    to inherit from this, and common methods can be added to it.
    '''

    def __init_subclass__(cls):
        cls.__full_path__ = f"{cls.__module__}.{cls.__qualname__}"

    def flash(self, zs=None, T=None, P=None, VF=None, SF=None, V=None, H=None,
              S=None, G=None, U=None, A=None, solution=None, hot_start=None,
              retry=False, dest=None, rho=None, rho_mass=None, H_mass=None,
              S_mass=None, G_mass=None, U_mass=None, A_mass=None,
              spec_fun=None, H_reactive=None):
        r'''Method to perform a flash calculation and return the result as an
        :obj:`EquilibriumState <thermo.equilibrium.EquilibriumState>` object.
        This generic interface allows flashes with any combination of valid
        specifications; if a flash is unimplemented and error will be raised.

        Parameters
        ----------
        zs : list[float], optional
            Mole fractions of each component, required unless there is only
            one component, [-]
        T : float, optional
            Temperature, [K]
        P : float, optional
            Pressure, [Pa]
        VF : float, optional
            Vapor fraction, [-]
        SF : float, optional
            Solid fraction, [-]
        V : float, optional
            Molar volume of the overall bulk, [m^3/mol]
        H : float, optional
            Molar enthalpy of the overall bulk, [J/mol]
        S : float, optional
            Molar entropy of the overall bulk, [J/(mol*K)]
        G : float, optional
            Molar Gibbs free energy of the overall bulk, [J/mol]
        U : float, optional
            Molar internal energy of the overall bulk, [J/mol]
        A : float, optional
            Molar Helmholtz energy of the overall bulk, [J/mol]
        solution : str or int, optional
           When multiple solutions exist, if more than one is found they will
           be sorted by T (and then P) increasingly; this number will index
           into the multiple solution array. Negative indexing is supported.
           'high' is an alias for 0, and 'low' an alias for -1. Setting this
           parameter may make a flash slower because in some cases more checks
           are performed. [-]
        hot_start : :obj:`EquilibriumState <thermo.equilibrium.EquilibriumState>`
            A previously converged flash or initial guessed state from which
            the flash can begin; this parameter can save time in some cases,
            [-]
        retry : bool
            Usually for flashes like UV or PH, there are multiple sets of
            possible iteration variables. For the UV case, the prefered
            iteration variable is P, so each iteration a PV solve is done on
            the phase; but equally the flash can be done iterating on
            `T`, where a TV solve is done on the phase each iteration.
            Depending on the tolerances, the flash type, the thermodynamic
            consistency of the phase, and other factors, it is possible the
            flash can fail. If `retry` is set to True, the alternate variable
            set will be iterated as a backup if the first flash fails. [-]
        dest : None or :obj:`EquilibriumState <thermo.equilibrium.EquilibriumState>` or :obj:`EquilibriumStream <thermo.stream.EquilibriumStream>`
            What type of object the flash result is set into; leave as None to
            obtain the normal `EquilibriumState` results, [-]
        rho : float, optional
            Molar density of the overall bulk; this is trivially converted to
            a `V` spec, [mol/m^3]
        rho_mass : float, optional
            Mass density of the overall bulk; this is trivially converted to
            a `rho` spec, [kg/m^3]
        H_mass : float, optional
            Mass enthalpy of the overall bulk; this is trivially converted to
            a `H` spec, [J/kg]
        S_mass : float, optional
            Mass entropy of the overall bulk; this is trivially converted to
            a `S` spec, [J/(kg*K)]
        G_mass : float, optional
            Mass Gibbs free energy of the overall bulk; this is trivially converted to
            a `G` spec, [J/kg]
        U_mass : float, optional
            Mass internal energy of the overall bulk; this is trivially converted to
            a `U` spec, [J/kg]
        A_mass : float, optional
            Mass Helmholtz energy of the overall bulk; this is trivially converted to
            a `A` spec, [J/kg]
        H_reactive : float, optional
            Molar reactive enthalpy, [J/mol]

        Returns
        -------
        results : :obj:`EquilibriumState <thermo.equilibrium.EquilibriumState>`
            Equilibrium object containing the state of the phases after the
            flash calculation [-]

        Notes
        -----
        .. warning::
            Not all flash specifications have a unique solution. Not all flash
            specifications will converge, whether from a bad model, bad inputs,
            or simply a lack of convergence by the implemented algorithms.
            You are welcome to submit these cases to the author but the library
            is provided AS IS, with NO SUPPORT.

        .. warning::
            Convergence of a flash may be impaired by providing `hot_start`.
            If reliability is desired, do not use this parameter.

        .. warning::
            The most likely thermodynamic methods to converge are
            thermodynamically consistent ones. This means e.g. an ideal liquid
            and an ideal gas; or an equation of state for both
            phases. Mixing thermodynamic models increases the possibility of
            multiple solutions, discontinuities, and other not-fun issues for
            the algorithms.

        Examples
        --------
        '''
        # print('flashing',T, P, H, S, VF)
        if zs is None:
            if self.N == 1:
                zs = [1.0] if self.scalar else ones(1)
            else:
                raise ValueError("Composition missing for flash")
        constants, correlations = self.constants, self.correlations
        settings = self.settings
        if dest is None:
            dest = EquilibriumState
#        if self.N > 1 and 0:
#            for zi in zs:
#                if zi == 1.0:
#                    # Does not work - phases expect multiple components mole fractions
#                    return self.flash_pure.flash(zs=zs, T=T, P=P, VF=VF, SF=SF,
#                                           V=V, H=H, S=S, U=U, G=G, A=A,
#                                           solution=solution, retry=retry,
#                                           hot_start=hot_start)
        if rho is not None:
            V, rho = 1.0/rho, None
        if rho_mass is not None:
            V, rho_mass = rho_to_Vm(rho_mass, mixing_simple(zs, constants.MWs)), None
        if H_mass is not None:
            H, H_mass = property_mass_to_molar(H_mass, mixing_simple(zs, constants.MWs)), None
        if S_mass is not None:
            S, S_mass = property_mass_to_molar(S_mass, mixing_simple(zs, constants.MWs)), None
        if G_mass is not None:
            G, G_mass = property_mass_to_molar(G_mass, mixing_simple(zs, constants.MWs)), None
        if U_mass is not None:
            U, U_mass = property_mass_to_molar(U_mass, mixing_simple(zs, constants.MWs)), None
        if A_mass is not None:
            A, A_mass = property_mass_to_molar(A_mass, mixing_simple(zs, constants.MWs)), None
        if H_reactive is not None:
            Hfgs = constants.Hfgs
            H = H_reactive - sum(zs[i]*Hfgs[i] for i in range(constants.N))
            H_reactive = None

        T_spec = T is not None
        P_spec = P is not None
        V_spec = V is not None
        H_spec = H is not None
        S_spec = S is not None
        U_spec = U is not None
        # Normally multiple solutions
        A_spec = A is not None
        G_spec = G is not None

        HSGUA_spec_count = H_spec + S_spec + G_spec + U_spec + A_spec


        VF_spec = VF is not None
        SF_spec = SF is not None

        flash_specs = {'zs': zs}
        if T_spec:
            T = float(T)
            flash_specs['T'] = T
            if T < self.T_MIN_FLASH_ANY:
                raise ValueError("Specified temperature ({} K) is below the minimum temeprature ({} K) "
                                 "supported by any of the provided phases".format(T, self.T_MIN_FLASH_ANY))
            # if T <= 0.0:
            #     raise ValueError("Specified temperature (%s K) is unphysical" %(T,))
        if P_spec:
            P = float(P)
            flash_specs['P'] = P
            if P <= 0.0:
                raise ValueError(f"Specified pressure ({P} Pa) is unphysical")
        if V_spec:
            # high-precision volumes are needed in some cases
            # V = float(V)
            flash_specs['V'] = V
            if V <= 0.0:
                raise ValueError(f"Specified molar volume ({V} m^3/mol) is unphysical")
        if H_spec:
            H = float(H)
            flash_specs['H'] = H
        if S_spec:
            S = float(S)
            flash_specs['S'] = S
        if U_spec:
            U = float(U)
            flash_specs['U'] = U
        if G_spec:
            G = float(G)
            flash_specs['G'] = G
        if A_spec:
            A = float(A)
            flash_specs['A'] = A

        if VF_spec:
            VF_spec = float(VF_spec)
            flash_specs['VF'] = VF
            if VF < 0.0 or VF > 1.0:
                raise ValueError(f"Specified vapor fraction ({VF}) is not between 0 and 1")
            elif not self.supports_VF_flash:
                raise ValueError("Cannot flash with a vapor fraction spec without at least one gas and liquid phase defined")
        if SF_spec:
            SF_spec = float(SF_spec)
            flash_specs['SF'] = SF
            if SF < 0.0 or SF > 1.0:
                raise ValueError(f"Specified solid fraction ({VF}) is not between 0 and 1")
            elif not self.supports_SF_flash:
                raise ValueError("Cannot flash with a solid fraction spec without at least one gas and liquid phase defined, as well as a solid phase")

        if ((T_spec and (P_spec or V_spec)) or (P_spec and V_spec)):
            g, ls, ss, betas, flash_convergence = self.flash_TPV(T=T, P=P, V=V, zs=zs, solution=solution, hot_start=hot_start)
            # TODO can creating a list here be avoided?
            if g is not None:
                id_phases = [g] + ls + ss
            else:
                id_phases = ls + ss

            g, ls, ss, betas = identify_sort_phases(id_phases, betas, constants,
                                                    correlations, settings=settings,
                                                    skip_solids=self.skip_solids)

            a_phase = id_phases[0]
            return dest(a_phase.T, a_phase.P, zs, gas=g, liquids=ls, solids=ss,
                                    betas=betas, flash_specs=flash_specs,
                                    flash_convergence=flash_convergence,
                                    constants=constants, correlations=correlations,
                                    settings=settings, flasher=self)

        elif T_spec and VF_spec:
            # All dew/bubble are the same with 1 component
            Psat, ls, g, iterations, err = self.flash_TVF(T, VF=VF, zs=zs, hot_start=hot_start)
            if type(ls) is not list:
                ls = [ls]
            flash_convergence = {'iterations': iterations, 'err': err}

            return dest(T, Psat, zs, gas=g, liquids=ls, solids=[],
                                    betas=[VF, 1.0 - VF], flash_specs=flash_specs,
                                    flash_convergence=flash_convergence,
                                    constants=constants, correlations=correlations,
                                    settings=settings, flasher=self)

        elif P_spec and VF_spec:
            # All dew/bubble are the same with 1 component
            Tsat, ls, g, iterations, err = self.flash_PVF(P, VF=VF, zs=zs, hot_start=hot_start)
            if type(ls) is not list:
                ls = [ls]
            flash_convergence = {'iterations': iterations, 'err': err}

            return dest(Tsat, P, zs, gas=g, liquids=ls, solids=[],
                                    betas=[VF, 1.0 - VF], flash_specs=flash_specs,
                                    flash_convergence=flash_convergence,
                                    constants=constants, correlations=correlations,
                                    settings=settings, flasher=self)
        elif T_spec and SF_spec:
            Psub, other_phase, s, iterations, err = self.flash_TSF(T, SF=SF, zs=zs, hot_start=hot_start)
            if other_phase.is_gas:
                g, liquids = other_phase, []
            else:
                g, liquids = None, [other_phase]
            flash_convergence = {'iterations': iterations, 'err': err}
#
            return dest(T, Psub, zs, gas=g, liquids=liquids, solids=[s],
                                    betas=[1-SF, SF], flash_specs=flash_specs,
                                    flash_convergence=flash_convergence,
                                    constants=constants, correlations=correlations,
                                    settings=settings, flasher=self)
        elif P_spec and SF_spec:
            Tsub, other_phase, s, iterations, err = self.flash_PSF(P, SF=SF, zs=zs, hot_start=hot_start)
            if other_phase.is_gas:
                g, liquids = other_phase, []
            else:
                g, liquids = None, [other_phase]
            flash_convergence = {'iterations': iterations, 'err': err}
#
            return dest(Tsub, P, zs, gas=g, liquids=liquids, solids=[s],
                                    betas=[1-SF, SF], flash_specs=flash_specs,
                                    flash_convergence=flash_convergence,
                                    constants=constants, correlations=correlations,
                                    settings=settings, flasher=self)
        elif VF_spec and any([H_spec, S_spec, U_spec, G_spec, A_spec]):
            spec_var, spec_val = [(k, v) for k, v in flash_specs.items() if k not in ('VF', 'zs')][0]
            T, Psat, liquid, gas, iters_inner, err_inner, err, iterations = self.flash_VF_HSGUA(VF, spec_val, fixed_var='VF', spec_var=spec_var, zs=zs, solution=solution, hot_start=hot_start)
            flash_convergence = {'iterations': iterations, 'err': err, 'inner_flash_convergence': {'iterations': iters_inner, 'err': err_inner}}
            return dest(T, Psat, zs, gas=gas, liquids=[liquid], solids=[],
                                    betas=[VF, 1.0 - VF], flash_specs=flash_specs,
                                    flash_convergence=flash_convergence,
                                    constants=constants, correlations=correlations,
                                    settings=settings, flasher=self)
        elif HSGUA_spec_count == 2:
            pass

        single_iter_key = (T_spec, P_spec, V_spec, H_spec, S_spec, U_spec)
        if single_iter_key in spec_to_iter_vars:
            fixed_var, spec, iter_var = spec_to_iter_vars[single_iter_key]
            _, _, iter_var_backup = spec_to_iter_vars_backup[single_iter_key]
            if T_spec:
                fixed_var_val = T
            elif P_spec:
                fixed_var_val = P
            else:
                fixed_var_val = V

            if H_spec:
                spec_val = H
            elif S_spec:
                spec_val = S
            else:
                spec_val = U

            # Only allow one
#            g, ls, ss, betas, flash_convergence = self.flash_TPV_HSGUA(fixed_var_val, spec_val, fixed_var, spec, iter_var)
            try:
                g, ls, ss, betas, flash_convergence = self.flash_TPV_HSGUA(fixed_var_val, spec_val, fixed_var, spec, iter_var, zs=zs, solution=solution, hot_start=hot_start, spec_fun=spec_fun)
            except Exception as e:
                if retry:
                    print('retrying HSGUA flash')
                    g, ls, ss, betas, flash_convergence = self.flash_TPV_HSGUA(fixed_var_val, spec_val, fixed_var, spec, iter_var_backup, zs=zs, solution=solution, hot_start=hot_start, spec_fun=spec_fun)
                else:
                    raise e
#            except UnconvergedError as e:
#                 if fixed_var == 'T' and iter_var in ('S', 'H', 'U'):
#                     g, ls, ss, betas, flash_convergence = self.flash_TPV_HSGUA(fixed_var_val, spec_val, fixed_var, spec, iter_var_backup, solution=solution)
#                 else:
                # Not sure if good idea - would prefer to converge without
            phases = ls + ss
            if g:
                phases += [g]
            T, P = phases[0].T, phases[0].P
            if self.N > 1:
                g, ls, ss, betas = identify_sort_phases(phases, betas, constants,
                                                        correlations, settings=settings,
                                                        skip_solids=self.skip_solids)

            return dest(T, P, zs, gas=g, liquids=ls, solids=ss,
                                    betas=betas, flash_specs=flash_specs,
                                    flash_convergence=flash_convergence,
                                    constants=constants, correlations=correlations,
                                    settings=settings, flasher=self)

        else:
            raise Exception('Flash inputs unsupported')

    flash_phase_boundary_algos = [flash_phase_boundary_one_sided_secant]
    flash_phase_boundary_methods = [SECANT_PHASE_BOUNDARY]

    FLASH_PHASE_BOUNDARY_MAXITER = 100
    FLASH_PHASE_BOUNDARY_MAXITER_XTOL = None
    FLASH_PHASE_BOUNDARY_MAXITER_YTOL = 1e-6


    def flash_phase_boundary(self, zs, phase_frac_check, T=None, P=None, V=None, H=None, S=None, U=None,
                             hot_start=None):
        T_spec = T is not None
        P_spec = P is not None
        V_spec = V is not None
        H_spec = H is not None
        S_spec = S is not None
        U_spec = U is not None
        spec_count = T_spec + P_spec + V_spec + H_spec + S_spec + U_spec
        if spec_count != 1:
            raise ValueError("One specification must be provided")
        if T_spec:
            iter_var, backup_iter_var = 'P', 'H'
            spec_var, spec_val = 'T', T
        if P_spec:
            iter_var, backup_iter_var = 'T', 'H'
            spec_var, spec_val = 'P', P
        if V_spec:
            iter_var, backup_iter_var = 'P', 'H'
            spec_var, spec_val = 'V', V
        if H_spec:
            iter_var, backup_iter_var = 'P', 'T'
            spec_var, spec_val = 'H', H
        if S_spec:
            iter_var, backup_iter_var = 'P', 'T'
            spec_var, spec_val = 'S', S
        if U_spec:
            iter_var, backup_iter_var = 'P', 'T'
            spec_var, spec_val = 'U', U


        for method in self.flash_phase_boundary_algos:
            res, bounding_attempts, iterations = method(flasher=self, zs=zs, spec_var=spec_var, spec_val=spec_val, iter_var=iter_var,
                                                    check=phase_frac_check, hot_start=hot_start,
                                                    xtol=self.FLASH_PHASE_BOUNDARY_MAXITER_XTOL,
                                                    ytol=self.FLASH_PHASE_BOUNDARY_MAXITER_YTOL,
                                                    maxiter=self.FLASH_PHASE_BOUNDARY_MAXITER)

            res.flash_convergence = {'inner_flash_convergence': res.flash_convergence,
                                     'bounding_attempts': bounding_attempts,
                                     'iterations': iterations}


            return res
        raise ValueError("Could not find a solution")



    flash_mixing_phase_boundary_methods = [PT_SECANT_PHASE_MIXING_BOUNDARY,
                                           NAIVE_BISECTION_PHASE_MIXING_BOUNDARY,
                                           SATURATION_SECANT_PHASE_MIXING_BOUNDARY,
                                           ]
    flash_mixing_phase_boundary_algos = [incipient_phase_one_sided_secant,
                                         incipient_phase_bounded_naive,
                                         incipient_liquid_bounded_PT_sat,
                                         ]

    FLASH_MIXING_PHASE_BOUNDARY_MAXITER = 100
    FLASH_MIXING_PHASE_BOUNDARY_XTOL = 1e-6
    FLASH_MIXING_PHASE_BOUNDARY_YTOL = 1e-6

    def flash_mixing_phase_boundary(self, specs, zs_existing, zs_added, boundary='VL'):
        if boundary == 'VL':
            # if we start from a liquid, will converge to a gas with VF=0
            # if we start from a gas, will converge to a liquid with LF=0
            check = VL_boolean_check
        elif boundary == 'LL':
            check = LL_boolean_check
        elif boundary == 'VLL':
            # Hard to say if this flash will converge to a VF=0 or a second liquid fraction of 0
            check = VLL_boolean_check
        elif boundary == 'VLN/LN':
            check = VLN_or_LN_boolean_check
        elif boundary == 'VLL/LL':
            check = VLL_or_LL_boolean_check
        else:
            raise ValueError("Unrecognized boundary")


        for method in self.flash_mixing_phase_boundary_algos:
            # try:
            if method is incipient_phase_one_sided_secant:
                # can only solve two phase prolems
                if boundary not in ('VL',):
                    continue
                res, bounding_attempts, iters, mixing_factor = incipient_phase_one_sided_secant(
                    flasher=self, specs=specs, zs_existing=zs_existing, zs_added=zs_added, check=check, ytol=self.FLASH_MIXING_PHASE_BOUNDARY_YTOL)
            elif method is incipient_phase_bounded_naive:
                res, bounding_attempts, iters, mixing_factor = incipient_phase_bounded_naive(flasher=self, specs=specs, zs_existing=zs_existing, zs_added=zs_added, check=check,
                                                                                    xtol=self.FLASH_MIXING_PHASE_BOUNDARY_XTOL)
            elif method is incipient_liquid_bounded_PT_sat and boundary == 'VL':
                res, bounding_attempts, iters, mixing_factor = incipient_liquid_bounded_PT_sat(flasher=self, specs=specs, zs_existing=zs_existing, zs_added=zs_added, check=check,
                                                                                            xtol=self.FLASH_MIXING_PHASE_BOUNDARY_XTOL)


            res.flash_convergence = {'inner_flash_convergence': res.flash_convergence,
                                     'bounding_attempts': bounding_attempts,
                                     'iterations': iters,
                                     'mixing_factor': mixing_factor}


            return res
            # except Exception as e:
            #     print(e)
        raise ValueError("Could not find a solution")



    def generate_Ts(self, Ts=None, Tmin=None, Tmax=None, pts=50, zs=None,
                    method=None):
        if method is None:
            method = 'physical'

        constants = self.constants

        N = constants.N
        if zs is None:
            zs = [1.0/N]*N

        physical = method == 'physical'
        realistic = method == 'realistic'

        Tcs = constants.Tcs
        Tc = sum([zs[i]*Tcs[i] for i in range(N)])

        if Ts is None:
            if Tmin is None:
                if physical:
                    Tmin = max(p.T_MIN_FIXED for p in self.phases)
                elif realistic:
                    # Round the temperature widely, ensuring consistent rounding
                    Tmin = 1e-2*round(floor(Tc), -1)
            if Tmax is None:
                if physical:
                    Tmax = min(p.T_MAX_FIXED for p in self.phases)
                elif realistic:
                    # Round the temperature widely, ensuring consistent rounding
                    Tmax = min(10*round(floor(Tc), -1), 2000)

            Ts = logspace(log10(Tmin), log10(Tmax), pts)
#            Ts = linspace(Tmin, Tmax, pts)
        return Ts


    def generate_Ps(self, Ps=None, Pmin=None, Pmax=None, pts=50, zs=None,
                    method=None):
        if method is None:
            method = 'physical'

        constants = self.constants

        N = constants.N
        if zs is None:
            zs = [1.0/N]*N

        physical = method == 'physical'
        realistic = method == 'realistic'

        Pcs = constants.Pcs
        Pc = sum([zs[i]*Pcs[i] for i in range(N)])

        if Ps is None:
            if Pmin is None:
                if physical:
                    Pmin = phases.Phase.P_MIN_FIXED
                elif realistic:
                    # Round the pressure widely, ensuring consistent rounding
                    Pmin = min(1e-5*round(floor(Pc), -1), 100)
            if Pmax is None:
                if physical:
                    Pmax = phases.Phase.P_MAX_FIXED
                elif realistic:
                    # Round the pressure widely, ensuring consistent rounding
                    Pmax = min(10*round(floor(Pc), -1), 1e8)

            Ps = logspace(log10(Pmin), log10(Pmax), pts)
        return Ps

    def generate_Vs(self, Vs=None, Vmin=None, Vmax=None, pts=50, zs=None,
                    method=None):
        if method is None:
            method = 'physical'

        constants = self.constants

        N = constants.N
        if zs is None:
            zs = [1.0/N]*N

        physical = method == 'physical'
        realistic = method == 'realistic'

        Vcs = constants.Vcs
        Vc = sum([zs[i]*Vcs[i] for i in range(N)])

        min_bound = None
        CEOS_phases = (phases.CEOSLiquid, phases.CEOSGas)
        for phase in self.phases:
            if isinstance(phase, CEOS_phases):
                c2R = phase.eos_class.c2*R
                Tcs, Pcs = constants.Tcs, constants.Pcs
                b = sum([c2R*Tcs[i]*zs[i]/Pcs[i] for i in range(constants.N)])
                min_bound = b*(1.0 + 1e-15) if min_bound is None else min(min_bound, b*(1.0 + 1e-15))



        if Vs is None:
            if Vmin is None:
                if physical:
                    Vmin = phases.Phase.V_MIN_FIXED
                elif realistic:
                    # Round the pressure widely, ensuring consistent rounding
                    Vmin = round(Vc, 5)
                if Vmin < min_bound:
                    Vmin = min_bound
            if Vmax is None:
                if physical:
                    Vmax = phases.Phase.V_MAX_FIXED
                elif realistic:
                    # Round the pressure widely, ensuring consistent rounding
                    Vmax = 1e5*round(Vc, 5)

            Vs = logspace(log10(Vmin), log10(Vmax), pts)
        return Vs

    def grid_flash(self, zs, Ts=None, Ps=None, Vs=None,
                   VFs=None, SFs=None, Hs=None, Ss=None, Us=None,
                   props=None, store=True):

        flashes = []

        T_spec = Ts is not None
        P_spec = Ps is not None
        V_spec = Vs is not None
        H_spec = Hs is not None
        S_spec = Ss is not None
        U_spec = Us is not None
        VF_spec = VFs is not None
        SF_spec = SFs is not None

        flash_specs = {'zs': zs}
        spec_keys = []
        spec_iters = []
        if T_spec:
            spec_keys.append('T')
            spec_iters.append(Ts)
        if P_spec:
            spec_keys.append('P')
            spec_iters.append(Ps)
        if V_spec:
            spec_keys.append('V')
            spec_iters.append(Vs)
        if H_spec:
            spec_keys.append('H')
            spec_iters.append(Hs)
        if S_spec:
            spec_keys.append('S')
            spec_iters.append(Ss)
        if U_spec:
            spec_keys.append('U')
            spec_iters.append(Us)
        if VF_spec:
            spec_keys.append('VF')
            spec_iters.append(VFs)
        if SF_spec:
            spec_keys.append('SF')
            spec_iters.append(SFs)

        do_props = props is not None
        scalar_props = isinstance(props, str)

        calc_props = []
        for n0, spec0 in enumerate(spec_iters[0]):
            if do_props:
                row_props = []
            if store:
                row_flashes = []
            for n1, spec1 in enumerate(spec_iters[1]):
                flash_specs = {'zs': zs, spec_keys[0]: spec0, spec_keys[1]: spec1}
                try:
                    state = self.flash(**flash_specs)
                except Exception as e:
                    state = None
                    print(f'Failed trying to flash {flash_specs}, with exception {e}.')

                if store:
                    row_flashes.append(state)
                if do_props:
                    if scalar_props:
                        state_props = state.value(props)if state is not None else None
                    else:
                        state_props = [state.value(s) for s in props] if state is not None else [None for s in props]

                    row_props.append(state_props)

            if do_props:
                calc_props.append(row_props)
            if store:
                flashes.append(row_flashes)

        if do_props and store:
            return flashes, calc_props
        elif do_props:
            return calc_props
        elif store:
            return flashes
        return None

    def _finish_initialization_base(self):
        solids = self.solids
        liquids = self.liquids
        gas = self.gas
        if solids is None:
            solids = []

        liquids_to_unique_liquids = []
        unique_liquids = []
        liquid_count = len(liquids)
        if liquid_count == 1 or (liquid_count > 1 and all(liquids[0].is_same_model(l) for l in liquids[1:])):
            # All the liquids are the same
            unique_liquids.append(liquids[0])
            liquids_to_unique_liquids.extend([0]*len(liquids))
        else:
            unique_liquid_hashes = []
            # unique_liquid_hashes is not used except in this code block
            for i, l in enumerate(liquids):
                h = l.model_hash()
                if h not in unique_liquid_hashes:
                    unique_liquid_hashes.append(h)
                    unique_liquids.append(l)
                    liquids_to_unique_liquids.append(i)
                else:
                    liquids_to_unique_liquids.append(unique_liquid_hashes.index(h))
        gas_to_unique_liquid = None
        if gas is not None:
            if len(unique_liquids) == 1:
                for i, l in enumerate(liquids):
                    if l.is_same_model(gas, ignore_phase=True):
                        gas_to_unique_liquid = liquids_to_unique_liquids[i]
                        self.ceos_gas_liquid_compatible = True
                        break

        self.gas_to_unique_liquid = gas_to_unique_liquid
        self.liquids_to_unique_liquids = liquids_to_unique_liquids

        self.unique_liquids = unique_liquids
        self.unique_liquid_count = len(unique_liquids)
        self.unique_phases = [gas] + unique_liquids if gas is not None else unique_liquids
        if solids:
            self.unique_phases += solids
        self.unique_phase_count = (1 if gas is not None else 0) + self.unique_liquid_count + len(solids)





        self.T_MIN_FLASH = max(p.T_MIN_FLASH for p in self.phases)
        self.T_MAX_FLASH = min(p.T_MAX_FLASH for p in self.phases)
        self.T_MIN_FLASH_ANY = min(p.T_MIN_FLASH for p in self.phases)
        self.T_MAX_FLASH_ANY = max(p.T_MAX_FLASH for p in self.phases)
        scalar = True
        scalar_statuses = {i.scalar for i in self.phases}
        if len(scalar_statuses) > 1:
            raise ValueError("Can only perform flashes with all phases in a numpy basis or all phases in a pure Python basis")
        self.scalar = scalar_statuses.pop()

        self.supports_lnphis_args = all(p.supports_lnphis_args for p in self.phases)
        
        # Make the phases aware of the constants and properties
        constants = self.constants
        correlations = self.correlations
        for p in self.phases:
            if hasattr(p, 'constants') and p.constants is not constants:
                raise ValueError("Provided phase is associated with a different constants object")
            p.constants = constants
            if hasattr(p, 'correlations') and p.correlations is not correlations:
                raise ValueError("Provided phase is associated with a different correlations object")
            p.correlations = correlations


    def debug_grid_flash(self, zs, check0, check1, Ts=None, Ps=None, Vs=None,
                         VFs=None, SFs=None, Hs=None, Ss=None, Us=None,
                         retry=False, verbose=True):

        matrix_spec_flashes = []
        matrix_flashes = []
        nearest_check_prop = 'T' if 'T' not in (check0, check1) else 'P'

        T_spec = Ts is not None
        P_spec = Ps is not None
        V_spec = Vs is not None
        H_spec = Hs is not None
        S_spec = Ss is not None
        U_spec = Us is not None
        VF_spec = VFs is not None
        SF_spec = SFs is not None

        flash_specs = {'zs': zs}
        spec_keys = []
        spec_iters = []
        if T_spec:
            spec_keys.append('T')
            spec_iters.append(Ts)
        if P_spec:
            spec_keys.append('P')
            spec_iters.append(Ps)
        if V_spec:
            spec_keys.append('V')
            spec_iters.append(Vs)
        if H_spec:
            spec_keys.append('H')
            spec_iters.append(Hs)
        if S_spec:
            spec_keys.append('S')
            spec_iters.append(Ss)
        if U_spec:
            spec_keys.append('U')
            spec_iters.append(Us)
        if VF_spec:
            spec_keys.append('VF')
            spec_iters.append(VFs)
        if SF_spec:
            spec_keys.append('SF')
            spec_iters.append(SFs)

        V_set = {check1, check0}
        TV_iter = V_set == {'T', 'V'}
        PV_iter = V_set == {'P', 'V'}
        high_prec_V = TV_iter or PV_iter

        for n0, spec0 in enumerate(spec_iters[0]):
            row = []
            row_flashes = []
            row_spec_flashes = []
            for n1, spec1 in enumerate(spec_iters[1]):

                flash_specs = {'zs': zs, spec_keys[0]: spec0, spec_keys[1]: spec1}
                state = self.flash(**flash_specs)

                check0_spec = getattr(state, check0)
                try:
                    check0_spec = check0_spec()
                except:
                    pass
                check1_spec = getattr(state, check1)
                try:
                    check1_spec = check1_spec()
                except:
                    pass

                kwargs = {}
                kwargs[check0] = check0_spec
                kwargs[check1] = check1_spec

                # TV_iter is important to always do
                if TV_iter:
                    kwargs['V'] = state.V_iter(force=False)
                kwargs['retry'] = retry
                kwargs['solution'] = lambda new: abs(new.value(nearest_check_prop) - state.value(nearest_check_prop))
                try:
                    new = self.flash(**kwargs)
                    if PV_iter:
                        # Do a check here on tolerance
                        err = abs((new.value(nearest_check_prop) - state.value(nearest_check_prop))/state.value(nearest_check_prop))
                        if err > 1e-8:
                            kwargs['V'] = state.V_iter(force=True)
                            new = self.flash(**kwargs)
                except Exception as e:
                    # Was it a precision issue? Some flashes can be brutal
                    if 'V' in kwargs:
                        try:
                             kwargs['V'] = state.V_iter(True)
                             new = self.flash(**kwargs)
                        except Exception as e2:
                            new = None
                            if verbose:
                                print(f'Failed trying to flash {kwargs}, from original point {flash_specs}, with exception {e}.')
                    else:
                        new = None
                        if verbose:
                            print(f'Failed trying to flash {kwargs}, from original point {flash_specs}, with exception {e}.')
                row_spec_flashes.append(state)
                row_flashes.append(new)

            matrix_spec_flashes.append(row_spec_flashes)
            matrix_flashes.append(row_flashes)
        return matrix_spec_flashes, matrix_flashes

    def debug_err_flash_grid(self, matrix_spec_flashes, matrix_flashes,
                             check, method='rtol', verbose=True):
        matrix = []
        N0 = len(matrix_spec_flashes)
        N1 = len(matrix_spec_flashes[0])
        for i in range(N0):
            row = []
            for j in range(N1):
                state = matrix_spec_flashes[i][j]
                new = matrix_flashes[i][j]

                act = getattr(state, check)
                try:
                    act = act()
                except:
                    pass

                if new is None:
                    err = 1.0
                else:
                    calc = getattr(new, check)
                    try:
                        calc = calc()
                    except:
                        pass
                    if method == 'rtol':
                        err = abs((act - calc)/act)

                if err > 1e-6 and verbose:
                    try:
                        print([matrix_flashes[i][j].T, matrix_spec_flashes[i][j].T])
                        print([matrix_flashes[i][j].P, matrix_spec_flashes[i][j].P])
                        print(matrix_flashes[i][j], matrix_spec_flashes[i][j])
                    except:
                        pass
                row.append(err)
            matrix.append(row)
        return matrix


    def TPV_inputs(self, spec0='T', spec1='P', check0='P', check1='V', prop0='T',
                   Ts=None, Tmin=None, Tmax=None,
                   Ps=None, Pmin=None, Pmax=None,
                   Vs=None, Vmin=None, Vmax=None,
                   VFs=None, SFs=None,
                   auto_range=None, zs=None, pts=50,
                   trunc_err_low=1e-15, trunc_err_high=None, plot=True,
                   show=True, color_map=None, retry=False, verbose=True):

        specs = []
        for a_spec in (spec0, spec1):
            if 'T' == a_spec:
                Ts = self.generate_Ts(Ts=Ts, Tmin=Tmin, Tmax=Tmax, pts=pts, zs=zs,
                                      method=auto_range)
                specs.append(Ts)
            elif 'P' == a_spec:
                Ps = self.generate_Ps(Ps=Ps, Pmin=Pmin, Pmax=Pmax, pts=pts, zs=zs,
                                  method=auto_range)
                specs.append(Ps)
            elif 'V' == a_spec:
                Vs = self.generate_Vs(Vs=Vs, Vmin=Vmin, Vmax=Vmax, pts=pts, zs=zs,
                                  method=auto_range)
                specs.append(Vs)
            elif 'VF' == a_spec:
                if VFs is None:
                    VFs = linspace(0, 1, pts)
                specs.append(VFs)
            elif 'SF' == a_spec:
                if SFs is None:
                    SFs = linspace(0, 1, pts)
                specs.append(SFs)

        specs0, specs1 = specs
        matrix_spec_flashes, matrix_flashes = self.debug_grid_flash(zs,
            check0=check0, check1=check1, Ts=Ts, Ps=Ps, Vs=Vs, VFs=VFs,
            retry=retry, verbose=verbose)

        errs = self.debug_err_flash_grid(matrix_spec_flashes,
            matrix_flashes, check=prop0, verbose=verbose)

        if plot:
            import matplotlib.pyplot as plt
            from matplotlib import cm
            from matplotlib.colors import LogNorm
#            plt.ioff()
            X, Y = np.meshgrid(specs0, specs1)
            z = np.array(errs).T
            fig, ax = plt.subplots()
            z[np.where(abs(z) < trunc_err_low)] = trunc_err_low
            if trunc_err_high is not None:
                z[np.where(abs(z) > trunc_err_high)] = trunc_err_high

            if color_map is None:
                color_map = cm.viridis

            # im = ax.pcolormesh(X, Y, z, cmap=cm.PuRd, norm=LogNorm())
            im = ax.pcolormesh(X, Y, z, cmap=color_map, norm=LogNorm(vmin=trunc_err_low, vmax=trunc_err_high))
            # im = ax.pcolormesh(X, Y, z, cmap=cm.viridis, norm=LogNorm(vmin=1e-7, vmax=1))
            cbar = fig.colorbar(im, ax=ax)
            cbar.set_label('Relative error')

            ax.set_yscale('log')
            ax.set_xscale('log')
            ax.set_xlabel(spec0)
            ax.set_ylabel(spec1)

            max_err = np.max(errs)
            if trunc_err_low is not None and max_err < trunc_err_low:
                max_err = 0
            if trunc_err_high is not None and max_err > trunc_err_high:
                max_err = trunc_err_high

            ax.set_title(f'{check0} {check1} validation of {prop0}; Reference flash {spec0} {spec1}; max err {max_err:.1e}')


            if show:
                plt.show()

            return matrix_spec_flashes, matrix_flashes, errs, fig
        return matrix_spec_flashes, matrix_flashes, errs

    def grid_props(self, spec0='T', spec1='P', prop='H',
                   Ts=None, Tmin=None, Tmax=None,
                   Ps=None, Pmin=None, Pmax=None,
                   Vs=None, Vmin=None, Vmax=None,
                   VFs=None, SFs=None,
                   auto_range=None, zs=None, pts=50, plot=True,
                   show=True, color_map=None):

        specs = []
        for a_spec in (spec0, spec1):
            if 'T' == a_spec:
                Ts = self.generate_Ts(Ts=Ts, Tmin=Tmin, Tmax=Tmax, pts=pts, zs=zs,
                                      method=auto_range)
                specs.append(Ts)
            if 'P' == a_spec:
                Ps = self.generate_Ps(Ps=Ps, Pmin=Pmin, Pmax=Pmax, pts=pts, zs=zs,
                                  method=auto_range)
                specs.append(Ps)
            if 'V' == a_spec:
                Vs = self.generate_Vs(Vs=Vs, Vmin=Vmin, Vmax=Vmax, pts=pts, zs=zs,
                                  method=auto_range)
                specs.append(Vs)
            if 'VF' == a_spec:
                if VFs is None:
                    VFs = linspace(0, 1, pts)
                specs.append(VFs)
            if 'SF' == a_spec:
                if SFs is None:
                    SFs = linspace(0, 1, pts)
                specs.append(SFs)

        specs0, specs1 = specs
        props = self.grid_flash(zs, Ts=Ts, Ps=Ps, Vs=Vs, VFs=VFs, props=prop, store=False)
#        props = []
#        pts_iter = range(pts)
#        for i in pts_iter:
#            row = []
#            for j in pts_iter:
#                flash = matrix_flashes[i][j]
#                try:
#                    v = getattr(flash, prop)
#                    try:
#                        v = v()
#                    except:
#                        pass
#                except:
#                    v = None
#                row.append(v)
#            props.append(row)

        if plot:
            import matplotlib.pyplot as plt
            from matplotlib import cm
            from matplotlib.colors import LogNorm
            X, Y = np.meshgrid(specs0, specs1)
            z = np.array(props).T
            fig, ax = plt.subplots()

            if color_map is None:
                color_map = cm.viridis

            if np.any(np.array(props) < 0):
                norm = None
            else:
                norm = LogNorm()
            im = ax.pcolormesh(X, Y, z, cmap=color_map, norm=norm) #
            cbar = fig.colorbar(im, ax=ax)
            cbar.set_label(prop)

            ax.set_yscale('log')
            ax.set_xscale('log')
            ax.set_xlabel(spec0)
            ax.set_ylabel(spec1)

#            ax.set_title()
            if show:
                plt.show()
            return props, fig

        return props

    def debug_mixing_phase_boundary_PT(self, zs, zs_mixing, Pmin=None, Pmax=None,
                                       Tmin=None, Tmax=None, pts=50,
                ignore_errors=True, values=False, verbose=False, show=False,
                T_pts=None, P_pts=None, Ts=None, Ps=None, boundary='VL'): # pragma: no cover
        if not has_matplotlib() and not values:
            raise Exception('Optional dependency matplotlib is required for plotting')
        if Pmin is None:
            Pmin = 1e2
        if Pmax is None:
            Pmax = min(self.constants.Pcs)
        if Tmin is None:
            Tmin = min(self.constants.Tms)*.9
        if Tmax is None:
            Tmax = max(self.constants.Tcs)*1.5
        if T_pts is None:
            T_pts = pts
        if P_pts is None:
            P_pts = pts

        Ts = self.generate_Ts(Ts=Ts, Tmin=Tmin, Tmax=Tmax, pts=T_pts, zs=zs)
        Ps = self.generate_Ps(Ps=Ps, Pmin=Pmin, Pmax=Pmax, pts=P_pts, zs=zs)

        matrix = []
        for T in Ts:
            row = []
            for P in Ps:
                try:
                    state = self.flash_mixing_phase_boundary(specs={'T': T, 'P': P}, zs_existing=zs, zs_added=zs_mixing, boundary=boundary)
                    row.append(state.flash_convergence['mixing_factor'])
                except Exception as e:
                    if verbose:
                        print([T, P, e])
                    if ignore_errors:
                        row.append(nan)
                    else:
                        raise e
            matrix.append(row)

        if values:
            return Ts, Ps, matrix

        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()

        Ts, Ps = np.meshgrid(Ts, Ps)
        im = ax.pcolormesh(Ts, Ps, matrix, cmap=plt.get_cmap('viridis'))
        cbar = fig.colorbar(im, ax=ax)
        cbar.set_label('Factor')

        ax.set_yscale('log')
        ax.set_xlabel('Temperature [K]')
        ax.set_ylabel('Pressure [Pa]')
        plt.title(f'PT flash mixing {boundary} boundary flashes, zs={zs}, zs_mixing={zs_mixing}')
        if show:
            plt.show()
        else:
            return fig

    def debug_PT(self, zs, Pmin=None, Pmax=None, Tmin=None, Tmax=None, pts=50,
                ignore_errors=True, values=False, verbose=False, show=False,
                T_pts=None, P_pts=None, Ts=None, Ps=None): # pragma: no cover
        if not has_matplotlib() and not values:
            raise Exception('Optional dependency matplotlib is required for plotting')
        if Pmin is None:
            Pmin = 1e4
        if Pmax is None:
            Pmax = min(self.constants.Pcs)
        if Tmin is None:
            Tmin = min(self.constants.Tms)*.9
        if Tmax is None:
            Tmax = max(self.constants.Tcs)*1.5
        if T_pts is None:
            T_pts = pts
        if P_pts is None:
            P_pts = pts

        Ts = self.generate_Ts(Ts=Ts, Tmin=Tmin, Tmax=Tmax, pts=T_pts, zs=zs)
        Ps = self.generate_Ps(Ps=Ps, Pmin=Pmin, Pmax=Pmax, pts=P_pts, zs=zs)

#        Ps = logspace(log10(Pmin), log10(Pmax), pts)
#        Ts = linspace(Tmin, Tmax, pts)

        matrix = []
        for T in Ts:
            row = []
            for P in Ps:
                try:
                    state = self.flash(T=T, P=P, zs=zs)
                    row.append(state.phase)
                except Exception as e:
                    if verbose:
                        print([T, P, e])
                    if ignore_errors:
                        row.append('F')
                    else:
                        raise e
            matrix.append(row)

        if values:
            return Ts, Ps, matrix


        regions = {'V': 1, 'L': 2, 'S': 3, 'VL': 4, 'LL': 5, 'VLL': 6,
                       'VLS': 7, 'VLLS': 8, 'VLLSS': 9, 'LLL': 10, 'VLLL': 11,
                       'VLLLL': 12, 'LLLL': 13, 'F': 0}

        used_regions = set()
        for row in matrix:
            for v in row:
                used_regions.add(v)

        region_keys = list(regions.keys())
        used_keys = [i for i in region_keys if i in used_regions]

        regions_keys = [n for _, n in sorted(zip([regions[i] for i in used_keys], used_keys))]
        used_values = [regions[i] for i in regions_keys]

        new_map = list(range(len(used_values)))
        new_map_trans = {i: j for i, j in zip(used_values, new_map)}

        dat = [[new_map_trans[regions[matrix[j][i]]] for j in range(pts)] for i in range(pts)]
#        print(dat)
        import matplotlib.pyplot as plt
        from matplotlib import colors
        fig, ax = plt.subplots()
        Ts, Ps = np.meshgrid(Ts, Ps)

        # need 3 more
        cmap = colors.ListedColormap(['y','b','r', 'g', 'c', 'm', 'k', 'fuchsia', 'gold', 'lime', 'indigo', 'cyan'][0:len(used_values)])

        vmax = len(used_values) - 1

#        ax.scatter(Ts,Ps, s=dat)
        im = ax.pcolormesh(Ts, Ps, dat, cmap=cmap, norm=colors.Normalize(vmin=0, vmax=vmax)) # , cmap=color_map, norm=LogNorm()
        cbar = fig.colorbar(im, ax=ax)
        cbar.set_label('Phase')
        cbar.ax.locator_params(nbins=len(used_values))
#        cbar = plt.colorbar()

#        cs = ax.contourf(Ts, Ps, dat, levels=list(sorted(regions.values())))
#        cbar = fig.colorbar(ax)

        # used_region_keys = regions_keys
        cbar.ax.set_yticklabels(regions_keys)

        # cbar.ax.set_yticklabels([n for _, n in sorted(zip(regions.values(), regions.keys()))])
#        cbar.ax.set_yticklabels(regions_keys)
#        ax.set_yscale('log')
        plt.yscale('log')
        plt.xlabel('System temperature, K')
        plt.ylabel('System pressure, Pa')
#        plt.imshow(dat, interpolation='nearest')
#        plt.legend(loc='best', fancybox=True, framealpha=0.5)
#        return fig, ax

        if len(zs) > 4:
            zs = '...'
        plt.title('PT system flashes, zs=%s' %zs)
        if show:
            plt.show()
        else:
            return fig


    def plot_TP(self, zs, Tmin=None, Tmax=None, pts=50, branches=None,
                ignore_errors=True, values=False, show=True, hot=False): # pragma: no cover
        r'''Method to create a plot of the phase envelope as can be calculated
        from a series of temperature & vapor fraction spec flashes. By default
        vapor fractions of 0 and 1 are plotted; additional vapor fraction
        specifications can be specified in the `branches` argument as a list.

        Parameters
        ----------
        zs : list[float]
            Mole fractions of the feed, [-]
        Tmin : float, optional
            Minimum temperature to begin the plot, [K]
        Tmax : float, optional
            Maximum temperature to end the plot, [K]
        pts : int, optional
           The number of points to calculated for each vapor fraction value,
           [-]
        branches : list[float], optional
            Extra vapor fraction values to plot, [-]
        ignore_errors : bool, optional
            Whether to fail on a calculation failure, or to ignore the bad
            point, [-]
        values : bool, optional
            If True, the calculated values will be returned instead of
            plotted, [-]
        show : bool, optional
            If True, the plot will be created and displayed; if False and `values`
            is False, the plot Figure object will be returned but not displayed;
            and if False and `values` is False, no plot will be created or shown
            [-]
        hot : bool, optional
            Whether to restart the next flash from the previous flash or not
            (intended to speed the call when True), [-]

        Returns
        -------
        Ts : list[float]
            Temperatures, [K]
        P_dews : list[float]
            Bubble point pressures at the evaluated points, [Pa]
        P_bubbles : list[float]
            Dew point pressures at the evaluated points, [Pa]
        branch_Ps : None or list[list[float]]
            Pressures which yield the equilibrium vapor fractions specified;
            formatted as [[P1_VFx, P2_VFx, ... Pn_VFx], ...,
            [P1_VFy, P2_VFy, ... Pn_VFy]], [Pa]
        '''
        if not has_matplotlib() and not values:
            raise Exception('Optional dependency matplotlib is required for plotting')
        if not Tmin:
            Tmin = min(self.constants.Tms)
        if not Tmax:
            Tmax = min(self.constants.Tcs)
        Ts = linspace(Tmin, Tmax, pts)
        P_dews = []
        P_bubbles = []
        branch = branches is not None
        if branch:
            branch_Ps = [[] for i in range(len(branches))]
        else:
            branch_Ps = None

        state_TVF0, state_TVF1 = None, None
        for T in Ts:
            if not hot:
                state_TVF0, state_TVF1 = None, None
            try:
                state_TVF0 = self.flash(T=T, VF=0.0, zs=zs, hot_start=state_TVF0)
                assert state_TVF0 is not None
                P_bubbles.append(state_TVF0.P)
            except Exception as e:
                if ignore_errors:
                    P_bubbles.append(None)
                else:
                    raise e
            try:
                state_TVF1 = self.flash(T=T, VF=1.0, zs=zs, hot_start=state_TVF1)
                assert state_TVF1 is not None
                P_dews.append(state_TVF1.P)
            except Exception as e:
                if ignore_errors:
                    P_dews.append(None)
                else:
                    raise e

            if branch:
                for VF, Ps in zip(branches, branch_Ps):
                    try:
                        state = self.flash(T=T, VF=VF, zs=zs)
                        Ps.append(state.P)
                    except Exception as e:
                        if ignore_errors:
                            Ps.append(None)
                        else:
                            raise e
        if values and not show:
            return Ts, P_dews, P_bubbles, branch_Ps

        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        plt.semilogy(Ts, P_dews, label='TP dew point curve')
        plt.semilogy(Ts, P_bubbles, label='TP bubble point curve')
        plt.xlabel('System temperature, K')
        plt.ylabel('System pressure, Pa')
        plt.title('PT system curve, zs=%s' %zs)
        if branch:
            for VF, Ps in zip(branches, branch_Ps):
                plt.semilogy(Ts, Ps, label='TP curve for VF=%s'%VF)
        plt.legend(loc='best')
        if show:
            plt.show()
        if values:
            return Ts, P_dews, P_bubbles, branch_Ps
        else:
            return fig


    def plot_PT(self, zs, Pmin=None, Pmax=None, pts=50, branches=[],
                ignore_errors=True, values=False, show=True, hot=False): # pragma: no cover
        r'''Method to create a plot of the phase envelope as can be calculated
        from a series of pressure & vapor fraction spec flashes. By default
        vapor fractions of 0 and 1 are plotted; additional vapor fraction
        specifications can be specified in the `branches` argument as a list.

        Parameters
        ----------
        zs : list[float]
            Mole fractions of the feed, [-]
        Pmin : float, optional
            Minimum pressure to begin the plot, [Pa]
        Pmax : float, optional
            Maximum pressure to end the plot, [Pa]
        pts : int, optional
           The number of points to calculated for each vapor fraction value,
           [-]
        branches : list[float], optional
            Extra vapor fraction values to plot, [-]
        ignore_errors : bool, optional
            Whether to fail on a calculation failure, or to ignore the bad
            point, [-]
        values : bool, optional
            If True, the calculated values will be returned instead of
            plotted, [-]
        show : bool, optional
            If True, the plot will be created and displayed; if False and `values`
            is False, the plot Figure object will be returned but not displayed;
            and if False and `values` is False, no plot will be created or shown
            [-]
        hot : bool, optional
            Whether to restart the next flash from the previous flash or not
            (intended to speed the call when True), [-]

        Returns
        -------
        Ps : list[float]
            Pressures, [Pa]
        T_dews : list[float]
            Bubble point temperatures at the evaluated points, [K]
        T_bubbles : list[float]
            Dew point temperatures at the evaluated points, [K]
        branch_Ts : None or list[list[float]]
            Temperatures which yield the equilibrium vapor fractions specified;
            formatted as [[T1_VFx, T2_VFx, ... Tn_VFx], ...,
            [T1_VFy, T2_VFy, ... Tn_VFy]], [k]
        '''
        if not has_matplotlib() and not values:
            raise Exception('Optional dependency matplotlib is required for plotting')
        if not Pmin:
            Pmin = 1e4
        if not Pmax:
            Pmax = min(self.constants.Pcs)
        Ps = logspace(log10(Pmin), log10(Pmax), pts)
        T_dews = []
        T_bubbles = []
        branch = bool(len(branches))
        if branch:
            branch_Ts = [[] for i in range(len(branches))]
        else:
            branch_Ts = None
        state_PVF0, state_PVF1 = None, None
        for P in Ps:
            if not hot:
                state_PVF0, state_PVF1 = None, None
            try:
                state_PVF0 = self.flash(P=P, VF=0, zs=zs, hot_start=state_PVF0)
                assert state_PVF0 is not None
                T_bubbles.append(state_PVF0.T)
            except Exception as e:
                if ignore_errors:
                    T_bubbles.append(None)
                else:
                    raise e
            try:
                state_PVF1 = self.flash(P=P, VF=1, zs=zs, hot_start=state_PVF1)
                assert state_PVF1 is not None
                T_dews.append(state_PVF1.T)
            except Exception as e:
                if ignore_errors:
                    T_dews.append(None)
                else:
                    raise e

            if branch:
                for VF, Ts in zip(branches, branch_Ts):
                    try:
                        state = self.flash(P=P, VF=VF, zs=zs)
                        Ts.append(state.T)
                    except Exception as e:
                        if ignore_errors:
                            Ts.append(None)
                        else:
                            raise e
        if values:
            return Ps, T_dews, T_bubbles, branch_Ts
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        plt.plot(Ps, T_dews, label='PT dew point curve')
        plt.plot(Ps, T_bubbles, label='PT bubble point curve')
        plt.xlabel('System pressure, Pa')
        plt.ylabel('System temperature, K')
        plt.title('PT system curve, zs=%s' %zs)
        if branch:
            for VF, Ts in zip(branches, branch_Ts):
                plt.plot(Ps, Ts, label='PT curve for VF=%s'%VF)
        plt.legend(loc='best')

        if show:
            plt.show()
        else:
            return fig

    def plot_ternary(self, T=None, P=None, scale=10): # pragma: no cover
        r'''Method to create a ternary plot of the system at either a specified
        temperature or pressure.

        Parameters
        ----------
        T : float, optional
            Temperature, [K]
        P : float, optional
            Pressure, [Pa]
        '''
        if not has_matplotlib():
            raise Exception('Optional dependency matplotlib is required for plotting')
        try:
            import ternary
        except:
            raise Exception('Optional dependency python-ternary is required for ternary plotting')
        if self.N != 3:
            raise Exception('Ternary plotting requires a mixture of exactly three components')

        is_T_spec = T is not None
        if not is_T_spec and P is None:
            raise ValueError("Either T or P must be specified")
        values = []

        cond = {'T': T} if is_T_spec else {'P': P}

        def dew_at_zs(zs):
            res = self.flash(zs=zs, VF=0, **cond)
            if is_T_spec:
                values.append(res.P)
                return res.P
            else:
                values.append(res.T)
                return res.T

        def bubble_at_zs(zs):
            res = self.flash(zs=zs, VF=1, **cond)
            if is_T_spec:
                return res.P
            else:
                return res.T

        import matplotlib as mpl
        import matplotlib.pyplot as plt
        axes_colors = {'b': 'g', 'l': 'r', 'r':'b'}
        ticks = [round(i / float(10), 1) for i in range(10+1)]

        fig, ax = plt.subplots(1, 3, gridspec_kw = {'width_ratios':[4, 4, 1]})
        ax[0].axis("off")
        ax[1].axis("off")
        ax[2].axis("off")
        names = self.constants.aliases

        for axis, f, i in zip(ax[0:2], [dew_at_zs, bubble_at_zs], [0, 1]):
            figure, tax = ternary.figure(ax=axis, scale=scale)
            figure.set_size_inches(12, 4)
            if not i:
                tax.heatmapf(f, boundary=True, colorbar=False, vmin=0)
            else:
                tax.heatmapf(f, boundary=True, colorbar=False, vmin=0, vmax=max(values))

            tax.boundary(linewidth=2.0)
            tax.left_axis_label(f"mole fraction {names[1]}", offset=0.16, color=axes_colors['l'])
            tax.right_axis_label(f"mole fraction {names[0]}", offset=0.16, color=axes_colors['r'])
            tax.bottom_axis_label(f"mole fraction {names[2]}", offset=0.16, color=axes_colors['b'])

            tax.ticks(ticks=ticks, axis='rlb', linewidth=1, clockwise=True,
                      axes_colors=axes_colors, offset=0.03, tick_formats="%.1f")

            tax.gridlines(multiple=scale/10., linewidth=2,
                          horizontal_kwargs={'color':axes_colors['b']},
                          left_kwargs={'color':axes_colors['l']},
                          right_kwargs={'color':axes_colors['r']},
                          alpha=0.5)

        norm = plt.Normalize(vmin=0, vmax=max(values))
        sm = plt.cm.ScalarMappable(cmap=plt.get_cmap('viridis'), norm=norm)
        sm._A = []
        cb = plt.colorbar(sm, ax=ax[2])
        text = 'Pressure, [Pa]' if is_T_spec else 'Temperature, [K]'
        cb.set_label(text, rotation=270, ha='center', va='center')
        cb.locator = mpl.ticker.LinearLocator(numticks=7)
        cb.formatter = mpl.ticker.ScalarFormatter()
        cb.formatter.set_powerlimits((0, 0))
        cb.update_ticks()
        # plt.tight_layout()
        if is_T_spec:
            text = "Bubble pressure vs composition (left) and dew pressure vs composition (right) at %s K" %T
        else:
            text = "Bubble temperature vs composition (left) and dew temperature vs composition (right) at %s Pa" %P
        fig.suptitle(text, fontsize=14)
        fig.subplots_adjust(top=0.85)
        plt.show()


    def plot_Txy(self, P, pts=30, ignore_errors=True, values=False, show=True): # pragma: no cover
        r'''Method to create a Txy plot for a binary system (holding pressure
        constant); the mole fraction of the first species is varied.

        Parameters
        ----------
        P : float
            Pressure for the plot, [Pa]
        pts : int, optional
           The number of points to calculated for each vapor fraction value,
           [-]
        ignore_errors : bool, optional
            Whether to fail on a calculation failure, or to ignore the bad
            point, [-]
        values : bool, optional
            If True, the calculated values will be returned instead of
            plotted, [-]
        show : bool, optional
            If True, the plot will be created and displayed; if False and `values`
            is False, the plot Figure object will be returned but not displayed;
            and if False and `values` is False, no plot will be created or shown
            [-]

        Returns
        -------
        fig : Figure
            Plot object, [-]
        z1 : list[float]
            Mole fractions of the first component at each point, [-]
        z2 : list[float]
            Mole fractions of the second component at each point, [-]
        T_dews : list[float]
            Bubble point temperatures at the evaluated points, [K]
        T_bubbles : list[float]
            Dew point temperatures at the evaluated points, [K]
        '''
        if not has_matplotlib() and values is not False:
            raise Exception('Optional dependency matplotlib is required for plotting')
        if self.N != 2:
            raise Exception('Txy plotting requires a mixture of exactly two components')
        z1 = linspace(0, 1, pts)
        z2 = [1.0 - zi for zi in z1]
        Ts_dew = []
        Ts_bubble = []

        for i in range(pts):
            try:
                res = self.flash(P=P, VF=0, zs=[z1[i], z2[i]])
                Ts_bubble.append(res.T)
            except Exception as e:
                if ignore_errors:
                    Ts_bubble.append(None)
                else:
                    raise e
            try:
                res = self.flash(P=P, VF=1, zs=[z1[i], z2[i]])
                Ts_dew.append(res.T)
            except Exception as e:
                if ignore_errors:
                    Ts_dew.append(None)
                else:
                    raise e
        if values:
            return z1, z2, Ts_dew, Ts_bubble
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        names = self.constants.aliases
        plt.title('Txy diagram at P=%s Pa' %P)
        plt.plot(z1, Ts_dew, label='Dew temperature, K')
        plt.plot(z1, Ts_bubble, label='Bubble temperature, K')
        plt.xlabel(f'Mole fraction {names[0]}')
        plt.ylabel('System temperature, K')
        plt.legend(loc='best')
        if show:
            plt.show()
        else:
            return fig

    def plot_Pxy(self, T, pts=30, ignore_errors=True, values=False, show=True): # pragma: no cover
        r'''Method to create a Pxy plot for a binary system (holding temperature
        constant); the mole fraction of the first species is varied.

        Parameters
        ----------
        T : float
            Temperature for the plot, [K]
        pts : int, optional
           The number of points to calculated for each vapor fraction value,
           [-]
        ignore_errors : bool, optional
            Whether to fail on a calculation failure, or to ignore the bad
            point, [-]
        values : bool, optional
            If True, the calculated values will be returned instead of
            plotted, [-]
        show : bool, optional
            If True, the plot will be created and displayed; if False and `values`
            is False, the plot Figure object will be returned but not displayed;
            and if False and `values` is False, no plot will be created or shown
            [-]

        Returns
        -------
        fig : Figure
            Plot object, [-]
        z1 : list[float]
            Mole fractions of the first component at each point, [-]
        z2 : list[float]
            Mole fractions of the second component at each point, [-]
        P_dews : list[float]
            Bubble point pressures at the evaluated points, [Pa]
        P_bubbles : list[float]
            Dew point pressures at the evaluated points, [Pa]
        '''
        if not has_matplotlib() and values is not False:
            raise Exception('Optional dependency matplotlib is required for plotting')
        if self.N != 2:
            raise Exception('Pxy plotting requires a mixture of exactly two components')
        z1 = linspace(0, 1, pts)
        z2 = [1.0 - zi for zi in z1]
        Ps_dew = []
        Ps_bubble = []
        names = self.constants.aliases

        for i in range(pts):
            try:
                res = self.flash(T=T, VF=0, zs=[z1[i], z2[i]])
                Ps_bubble.append(res.P)
            except Exception as e:
                if ignore_errors:
                    Ps_bubble.append(None)
                else:
                    raise e
            try:
                res = self.flash(T=T, VF=1, zs=[z1[i], z2[i]])
                Ps_dew.append(res.P)
            except Exception as e:
                if ignore_errors:
                    Ps_dew.append(None)
                else:
                    raise e
        if values:
            return z1, z2, Ps_dew, Ps_bubble

        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        plt.title('Pxy diagram at T=%s K' %T)
        plt.plot(z1, Ps_dew, label='Dew pressure')
        plt.plot(z1, Ps_bubble, label='Bubble pressure')
        plt.xlabel(f'Mole fraction {names[0]}')
        plt.ylabel('System pressure, Pa')
        plt.legend(loc='best')
        if show:
            plt.show()
        else:
            return fig

    def plot_xy(self, P=None, T=None, pts=30, ignore_errors=True, values=False,
                show=True, VF=0.0): # pragma: no cover
        r'''Method to create a xy diagram for a binary system. Either a
        temperature or pressure can be specified. By default, bubble point
        flashes are performed; this can be varied by changing `VF`.

        Parameters
        ----------
        P : float, optional
            The specified pressure, [Pa]
        T : float, optional
            The specified temperature, [K]
        pts : int, optional
           The number of points in the plot [-]
        ignore_errors : bool, optional
            Whether to fail on a calculation failure, or to ignore the bad
            point, [-]
        values : bool, optional
            If True, the calculated values will be returned instead of
            plotted, [-]
        show : bool, optional
            If True, the plot will be created and displayed; if False and `values`
            is False, the plot Figure object will be returned but not displayed;
            and if False and `values` is False, no plot will be created or shown
            [-]

        Returns
        -------
        fig : Figure
            Plot object, [-]
        z1 : list[float]
            Overall mole fractions of the first component at each point, [-]
        z2 : list[float]
            Overall mole fractions of the second component at each point, [-]
        x1 : list[float]
            Liquid mole fractions of component 1 at each point, [-]
        y1 : list[float]
            Vapor mole fractions of component 1 at each point, [-]
        '''
        if not has_matplotlib():
            raise Exception('Optional dependency matplotlib is required for plotting')
        if self.N != 2:
            raise Exception('xy plotting requires a mixture of exactly two components')
        z1 = linspace(0.0, 1.0, pts)
        z2 = [1.0 - zi for zi in z1]
        y1_bubble = []
        x1_bubble = []
        for i in range(pts):
            try:
                if T is not None:
                    res = self.flash(T=T, VF=VF, zs=[z1[i], z2[i]])
                elif P is not None:
                    res = self.flash(P=P, VF=VF, zs=[z1[i], z2[i]])
                x1_bubble.append(res.liquid_bulk.zs[0])
                y1_bubble.append(res.gas.zs[0])
            except Exception as e:
                if ignore_errors:
                    print('Failed on pt %d' %(i), e)
                else:
                    raise e
        if values:
            return z1, z2, x1_bubble, y1_bubble
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        if T is not None:
            plt.title('xy diagram at T=%s K (varying P)' %T)
        else:
            plt.title('xy diagram at P=%s Pa (varying T)' %P)
        names = self.constants.aliases
        plt.xlabel(f'Liquid mole fraction {names[0]}')
        plt.ylabel(f'Vapor mole fraction {names[0]}')
        plt.plot(x1_bubble, y1_bubble, '-', label='liquid vs vapor composition')
        plt.legend(loc='best')
        plt.plot([0, 1], [0, 1], '--')
        plt.axis((0,1,0,1))
        if show:
            plt.show()
        else:
            return fig

    def V_liquids_ref(self):
        r'''Method to calculate and return the liquid reference molar volumes
        according to the temperature variable `T_liquid_volume_ref` of
        :obj:`thermo.bulk.BulkSettings`.

        Returns
        -------
        V_liquids_ref : list[float]
            Liquid molar volumes at the reference condition, [m^3/mol]

        Notes
        -----
        '''
        T_liquid_volume_ref = self.settings.T_liquid_volume_ref
        if T_liquid_volume_ref == 298.15:
            Vls = self.constants.Vml_STPs
        elif T_liquid_volume_ref == 288.7055555555555:
            Vls = self.constants.Vml_60Fs
        else:
            Vls = [i(T_liquid_volume_ref, None) for i in self.correlations.VolumeLiquids]
        return Vls


    @property
    def water_index(self):
        r'''The index of the component water in the components. None if water
        is not present. Water is recognized by its CAS number.

        Returns
        -------
        water_index : int
            The index of the component water, [-]

        Notes
        -----
        '''
        try:
            return self._water_index
        except AttributeError:
            pass

        try:
            self._water_index = self.constants.CASs.index(CAS_H2O)
        except ValueError:
            self._water_index = None
        return self._water_index



'''Chemical Engineering Design Library (ChEDL). Utilities for process modeling.
Copyright (C) 2016, 2017, 2018, 2019, 2020, 2021 Caleb Bell <Caleb.Andrew.Bell@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

import os

from fluids import numerics

if not numerics.is_micropython:

    from chemicals import (
        acentric,
        combustion,
        critical,
        dipole,
        dippr,
        elements,
        environment,
        flash_basic,
        identifiers,
        lennard_jones,
        miscdata,
        rachford_rice,
        reaction,
        refractivity,
        safety,
        solubility,
        temperature,
        triple,
        virial,
    )
    from chemicals.acentric import *  # noqa: F403
    from chemicals.combustion import *  # noqa: F403
    from chemicals.critical import *  # noqa: F403
    from chemicals.dipole import *  # noqa: F403
    from chemicals.dippr import *  # noqa: F403
    from chemicals.elements import *  # noqa: F403
    from chemicals.environment import *  # noqa: F403
    from chemicals.flash_basic import *  # noqa: F403
    from chemicals.identifiers import *  # noqa: F403
    from chemicals.lennard_jones import *  # noqa: F403
    from chemicals.miscdata import *  # noqa: F403
    from chemicals.rachford_rice import *  # noqa: F403
    from chemicals.reaction import *  # noqa: F403
    from chemicals.refractivity import *  # noqa: F403
    from chemicals.safety import *  # noqa: F403
    from chemicals.solubility import *  # noqa: F403
    from chemicals.temperature import *  # noqa: F403
    from chemicals.triple import *  # noqa: F403
    from chemicals.virial import *  # noqa: F403

    from . import (
        activity,
        bulk,
        chemical,
        chemical_package,
        chemical_utils,
        coolprop,
        datasheet,
        electrochem,
        eos,
        eos_alpha_functions,
        eos_mix,
        eos_mix_methods,
        eos_volume,
        equilibrium,
        fitting,
        flash,
        functional_groups,
        group_contribution,
        heat_capacity,
        interaction_parameters,
        interface,
        law,
        mixture,
        nrtl,
        permittivity,
        phase_change,
        phase_identification,
        phases,
        property_package,
        regular_solution,
        stream,
        thermal_conductivity,
        unifac,
        uniquac,
        utils,
        vapor_pressure,
        viscosity,
        volume,
        wilson,
        redlich_kister,
    )
    from .activity import *  # noqa: F403
    from .bulk import *  # noqa: F403
    from .chemical import *  # noqa: F403
    from .chemical_package import *  # noqa: F403
    from .chemical_utils import *  # noqa: F403
    from .coolprop import *  # noqa: F403
    from .datasheet import *  # noqa: F403
    from .electrochem import *  # noqa: F403
    from .eos import *  # noqa: F403
    from .eos_alpha_functions import *  # noqa: F403
    from .eos_mix import *  # noqa: F403
    from .eos_mix_methods import *  # noqa: F403
    from .eos_volume import *  # noqa: F403
    from .equilibrium import *  # noqa: F403
    from .fitting import *  # noqa: F403
    from .flash import *  # noqa: F403
    from .functional_groups import *  # noqa: F403
    from .group_contribution import *  # noqa: F403
    from .heat_capacity import *  # noqa: F403
    from .interaction_parameters import *  # noqa: F403
    from .interface import *  # noqa: F403
    from .law import *  # noqa: F403
    from .mixture import *  # noqa: F403
    from .nrtl import *  # noqa: F403
    from .permittivity import *  # noqa: F403
    from .phase_change import *  # noqa: F403
    from .phase_identification import *  # noqa: F403
    from .phases import *  # noqa: F403
    from .property_package import *  # noqa: F403
    from .regular_solution import *  # noqa: F403
    from .stream import *  # noqa: F403
    from .thermal_conductivity import *  # noqa: F403
    from .unifac import *  # noqa: F403
    from .uniquac import *  # noqa: F403
    from .utils import *  # noqa: F403
    from .vapor_pressure import *  # noqa: F403
    from .viscosity import *  # noqa: F403
    from .volume import *  # noqa: F403
    from .wilson import *  # noqa: F403
    from .redlich_kister import * # noqa: F403

    #from chemicals import *


    __all__ = ['rachford_rice', 'flash_basic', 'chemical', 'chemical_package', 'combustion', 'critical', 'flash',
     'dipole', 'electrochem', 'elements', 'environment', 'eos', 'eos_mix',
     'heat_capacity',  'identifiers', 'group_contribution', 'law', 'lennard_jones',
     'miscdata',
     'permittivity', 'phase_change', 'phases', 'property_package', 'reaction',
     'refractivity', 'safety', 'solubility', 'interface', 'interaction_parameters',
     'thermal_conductivity', 'triple', 'utils',
     'vapor_pressure', 'virial', 'viscosity', 'volume', 'acentric', 'coolprop',
     'datasheet', 'dippr', 'unifac', 'stream', 'mixture',
     'chemical_utils', 'wilson', 'nrtl', 'uniquac', 'regular_solution',
     'equilibrium', 'phase_identification', 'temperature', 'fitting',
     'eos_alpha_functions', 'eos_volume', 'bulk', 'eos_mix_methods', 'activity',
     'functional_groups', 'redlich_kister']

    __all__.extend(eos_volume.__all__)
    __all__.extend(eos_alpha_functions.__all__)
    __all__.extend(acentric.__all__)
    __all__.extend(rachford_rice.__all__)
    __all__.extend(flash_basic.__all__)
    __all__.extend(chemical_package.__all__)
    __all__.extend(chemical.__all__)
    __all__.extend(combustion.__all__)
    __all__.extend(critical.__all__)
    __all__.extend(coolprop.__all__)
    #__all__.extend(dipole.__all__)
    __all__.extend(dippr.__all__)
    __all__.extend(datasheet.__all__)
    __all__.extend(electrochem.__all__)
    __all__.extend(elements.__all__)
    __all__.extend(environment.__all__)
    __all__.extend(eos.__all__)
    __all__.extend(eos_mix.__all__)
    __all__.extend(flash.__all__)
    __all__.extend(heat_capacity.__all__)
    __all__.extend(identifiers.__all__)
    __all__.extend(interaction_parameters.__all__)
    __all__.extend(group_contribution.__all__)
    __all__.extend(law.__all__)
    __all__.extend(lennard_jones.__all__)
    __all__.extend(miscdata.__all__)
    __all__.extend(mixture.__all__)
    __all__.extend(permittivity.__all__)
    __all__.extend(phase_change.__all__)
    __all__.extend(phases.__all__)
    __all__.extend(phase_identification.__all__)
    __all__.extend(property_package.__all__)
    __all__.extend(reaction.__all__)
    __all__.extend(refractivity.__all__)
    __all__.extend(safety.__all__)
    __all__.extend(solubility.__all__)
    __all__.extend(stream.__all__)
    __all__.extend(interface.__all__)
    __all__.extend(thermal_conductivity.__all__)
    __all__.extend(triple.__all__)
    __all__.extend(utils.__all__)
    __all__.extend(unifac.__all__)
    __all__.extend(vapor_pressure.__all__)
    __all__.extend(virial.__all__)
    __all__.extend(viscosity.__all__)
    __all__.extend(volume.__all__)
    __all__.extend(chemical_utils.__all__)
    __all__.extend(wilson.__all__)
    __all__.extend(nrtl.__all__)
    __all__.extend(uniquac.__all__)
    __all__.extend(regular_solution.__all__)
    __all__.extend(equilibrium.__all__)
    __all__.extend(temperature.__all__)
    __all__.extend(bulk.__all__)
    __all__.extend(eos_mix_methods.__all__)
    __all__.extend(activity.__all__)
    __all__.extend(fitting.__all__)
    __all__.extend(functional_groups.__all__)
    __all__.extend(redlich_kister.__all__)

    # backwards compatibility hack to allow thermo.chemical.Mixture to still be importable
    try:
        chemical.__dict__['Mixture'] = mixture.Mixture
        chemical.__dict__['Stream'] = stream.Stream
    except:
        pass
    # However, they cannot go in thermo.chemical's __all__ or they will appear in the
    # documentation and Sphinx currently has no wat to exclude them
    submodules = [activity, chemical, chemical_package, chemical_utils, coolprop, datasheet,
                  electrochem, eos, eos_mix, equilibrium, heat_capacity,
                  identifiers, interaction_parameters, interface, group_contribution.joback, law,
                  mixture, nrtl, permittivity, phase_change, phase_identification,
                  property_package, regular_solution,
                  stream, thermal_conductivity, unifac, uniquac, safety,
                  fitting,functional_groups,
                  utils, vapor_pressure, viscosity, volume, wilson, eos_alpha_functions,
                  eos_volume, eos_mix_methods,
                  flash, flash.flash_base, flash.flash_pure_vls,
                  flash.flash_utils, flash.flash_vl, flash.flash_vln,
                  phases, phases.air_phase, phases.ceos, phases.combined,
                  phases.coolprop_phase, phases.gibbs_excess, phases.helmholtz_eos,
                  phases.iapws_phase, phases.ideal_gas, phases.petroleum,
                  phases.phase, phases.phase_utils, phases.virial_phase,
                  utils.functional, utils.mixture_property,
                  utils.t_dependent_property, utils.tp_dependent_property,
                  redlich_kister]

    def complete_lazy_loading():
        import chemicals
        chemicals.complete_lazy_loading()
        electrochem._load_electrochem_data()
        interaction_parameters.IPDB
        law.load_law_data()
        law.load_economic_data()
        unifac.load_unifac_ip()
        unifac.load_group_assignments_DDBST()
        try:
            pass
        except:
            pass
    if hasattr(os, '_called_from_test'):
        # pytest timings are hard to measure with lazy loading
        complete_lazy_loading()

    global vectorized, numba, units, numba_vectorized
    if numerics.PY37:
        def __getattr__(name):
            global vectorized, numba, units, numba_vectorized
            if name == 'vectorized':
                import thermo.vectorized
                return thermo.vectorized
            if name == 'numba':
                import thermo.numba
                return thermo.numba
            if name == 'units':
                import thermo.units
                return thermo.units
            if name == 'numba_vectorized':
                import thermo.numba_vectorized
                return thermo.numba_vectorized
            raise AttributeError(f"module {__name__} has no attribute {name}")
    else:
        pass

try:
    thermo_dir = os.path.dirname(__file__)
except:
    thermo_dir = ''

__version__ = '0.2.27'





'''Chemical Engineering Design Library (ChEDL). Utilities for process modeling.
Copyright (C) 2019, Caleb Bell <Caleb.Andrew.Bell@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

from math import log10

import numpy as np
from fluids.numerics import assert_close, assert_close1d, assert_close2d, assert_close3d


def check_np_output_activity(model, modelnp, modelnp2):
    # model is flat, scalar, list-based model
    # modelnp is numba model
    # modelnp2 is created from the numba model with to_T_xs at a different composition

    scalar_attrs = ['d3GE_dT3', 'd2GE_dT2', 'GE', 'dGE_dT']
    for attr in scalar_attrs:
        if hasattr(model, attr):
#            print(attr)
            assert_close(getattr(model, attr)(), getattr(modelnp, attr)(), rtol=2e-13)
            assert_close(getattr(modelnp2, attr)(), getattr(modelnp, attr)(), rtol=2e-13)
            assert type(getattr(model, attr)()) is float
    #        assert type(getattr(modelnp, attr)()) is float
    #        assert type(getattr(modelnp2, attr)()) is float

    vec_attrs = ['dGE_dxs', 'gammas', '_gammas_dGE_dxs',
                 'd2GE_dTdxs', 'dHE_dxs', 'gammas_infinite_dilution', 'dHE_dns',
                'dnHE_dns', 'dSE_dxs', 'dSE_dns', 'dnSE_dns', 'dGE_dns', 'dnGE_dns', 'd2GE_dTdns',
                'd2nGE_dTdns', 'dgammas_dT']

    for attr in vec_attrs:
#        print(attr)
        assert_close1d(getattr(model, attr)(), getattr(modelnp, attr)(), rtol=2e-13)
        assert_close1d(getattr(modelnp2, attr)(), getattr(modelnp, attr)(), rtol=2e-13)
        assert type(getattr(model, attr)()) is list
        assert type(getattr(modelnp, attr)()) is np.ndarray
        assert type(getattr(modelnp2, attr)()) is np.ndarray

    mat_attrs = ['d2GE_dxixjs', 'd2nGE_dninjs', 'dgammas_dns']
    for attr in mat_attrs:
#        print(attr)
        assert_close2d(getattr(model, attr)(), getattr(modelnp, attr)(), rtol=1e-12)
        assert_close2d(getattr(modelnp2, attr)(), getattr(modelnp, attr)(), rtol=1e-12)
        assert type(getattr(model, attr)()) is list
        assert type(getattr(modelnp, attr)()) is np.ndarray
        assert type(getattr(modelnp2, attr)()) is np.ndarray

    attrs_3d = ['d3GE_dxixjxks']
    for attr in attrs_3d:
        if hasattr(model, attr):
#            print(attr)
            # some models do not have this implemented
            assert_close3d(getattr(model, attr)(), getattr(modelnp, attr)(), rtol=1e-13)
            assert_close3d(getattr(modelnp2, attr)(), getattr(modelnp, attr)(), rtol=1e-13)
            assert type(getattr(model, attr)()) is list
            assert type(getattr(modelnp, attr)()) is np.ndarray
            assert type(getattr(modelnp2, attr)()) is np.ndarray


def plot_unsupported(reason, color='r'):
    '''Helper function - draw a plot with an `x` over it displaying a message
    why that plot is not supported.
    '''
    import matplotlib.pyplot as plt
    fig, ax = plt.subplots()

    xlims = ax.get_xlim()
    ylims = ax.get_ylim()
    ax.plot([0, 1], [0, 1], lw=5, c=color)
    ax.plot([0, 1], [1, 0], lw=5, c=color)

    ax.text(.5, .5, reason, ha='center', va='center', bbox=dict(fc='white'))
    return fig



def mark_plot_unsupported(plot_fig, reason, color='r'):
    ax = plot_fig.axes[0]
    xlims = ax.get_xlim()
    ylims = ax.get_ylim()
    xmid = 10**(0.5*(log10(xlims[0]) + log10(xlims[1])))
    ymid = 10**(0.5*(log10(ylims[0]) + log10(ylims[1])))
    ax.text(xmid, ymid, reason, ha='center', va='center', bbox=dict(fc='white'))
    color = 'r'
    ax.plot(xlims, ylims, lw=5, c=color)
    ax.plot(xlims, ylims[::-1], lw=5, c=color)



def flash_rounding(x):
    if isinstance(x, float):
        return float('%.10e' %(x))
    return x



'''Chemical Engineering Design Library (ChEDL). Utilities for process modeling.
Copyright (C) 2022 Caleb Bell <Caleb.Andrew.Bell@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''
__all__ = ['GibbsEOS',]

from thermo.phases.phase import Phase


class GibbsEOS(Phase):
    # http://www.iapws.org/relguide/Advise3.pdf is useful
    def V(self):    
        return self.dG_dP()
        # return self._V

    def dV_dP(self):
        # easy
        '''from sympy import *
        T, P = symbols('T, P')
        G = symbols('G', cls=Function)
        V = diff(G(T, P), P)
        diff(V, P)'''
        return self.d2G_dP2()

    dV_dP_T = dV_dP
    
    def dP_dV(self):
        return 1.0/self.d2G_dP2()

    dP_dV_T = dP_dV

    def d2V_dP2(self):
        return self.d3G_dP3()

    def dV_dT(self):
        return self.d2G_dTdP()

    def dT_dV(self):
        return 1.0/self.d2G_dTdP()
    
    dT_dV_P = dT_dV

    def d2V_dT2(self):
        return self.d3G_dT2dP()
    
    def d2V_dTdP(self):
        return self.d3G_dTdP2()


    def dP_dT(self):
        return -self.dV_dT()/self.dV_dP()

    dP_dT_V = dP_dT

    def dT_dP(self):
        return 1.0/self.dP_dT()
    
    dT_dP_V = dT_dP


    def H(self):
        return self.G() - self.T*self.dG_dT()

    def S(self):
        return -self.dG_dT()
    
    def Cp(self):
        return -self.T*self.d2G_dT2()
    
    def U(self):
        return self.G() - self.T*self.dG_dT() - self.P*self.dG_dP()

    def A(self):
        return self.G() - self.P*self.dG_dP()

    def dS_dT(self):
        return -self.d2G_dT2()
    dS_dT_P = dS_dT

    def d2S_dT2(self):
        return -self.d3G_dT3()
    d2S_dT2_P = d2S_dT2

    def dS_dP(self):
        return -self.d2G_dTdP()

    dS_dP_T = dS_dP

    def d2S_dP2(self):
        return -self.d3G_dTdP2()

    def d2S_dTdP(self):
        return -self.d3G_dT2dP()
    

    



'''Chemical Engineering Design Library (ChEDL). Utilities for process modeling.
Copyright (C) 2023 Caleb Bell <Caleb.Andrew.Bell@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

This module contains a class :obj:`NRTL` for performing activity coefficient
calculations with the NRTL model. An older, functional calculation for
activity coefficients only is also present, :obj:`NRTL_gammas`.

For reporting bugs, adding feature requests, or submitting pull requests,
please use the `GitHub issue tracker <https://github.com/CalebBell/thermo/>`_.

.. contents:: :local:
'''

__all__ = ['redlich_kister_reverse','redlich_kister_reverse_2d', 'redlich_kister_excess_inner',
'redlich_kister_build_structure', 'redlich_kister_T_dependence', 'redlich_kister_excess_inner_binary',
'redlich_kister_excess_binary', 'redlich_kister_fitting_to_use']
from math import log

def redlich_kister_reverse_2d(data_2d):
    data_2d = [d.copy() for d in data_2d]
    params = len(data_2d)
    T_dep_params = len(data_2d[0])
    for i, k in enumerate(range(params)):
        if k%2 == 1:
            for j in range(T_dep_params):
                data_2d[i][j] = -data_2d[i][j]
    return data_2d


def redlich_kister_reverse(data_1d):
    data_1d = data_1d.copy()
    params = len(data_1d)
    for i, k in enumerate(range(params)):
        if k%2 == 1:
            data_1d[i] = -data_1d[i]
    return data_1d


def redlich_kister_excess_inner(N, N_terms, a_tensor, xs):
    r'''Calculate the excess property for a given set of composition
    and temperature-independent constants.

    .. math::
        E = 0.5 \sum_i^N \sum_j^N \sum_k^{N_terms} \A_{ijk} x_i x_j (x_i - x_j)^k

    Parameters
    ----------
    N : int
        The number of components in the mixture, [-]
    N_terms : int
        The number of terms in the redlich-kister expansion. This must
        be the same for each component interaction; pad with zeros, [-]
    a_tensor : list[list[list[float]]]
        RK parameters, indexed by [i][j][k], [-]
    xs : list[float]
        Mole fractions, [-]

    Returns
    -------
    excess : float
        The calculated excess value; in some cases this is
        dimensionless and multiplied by a constant like `R*T`
        or `1000 kg/m^3` to obtain its dimensionality and
        in other cases it has the dimensions of the property
        being calculated like `N/m` or `Pa*s`, [-]

    Notes
    -----
    Temperature dependent constants should be calculated
    in an outer function.

    '''
    excess = 0.0
    for i in range(N):
        outer = 0.0
        for j in range(N):
#             if i ==j:
#                 continue
            inner = 0.0
            factor = 1.0
            diff = (xs[i] - xs[j])
            for k in range(0, N_terms):
                inner += a_tensor[i][j][k]*factor
                factor *= diff
            outer += inner*xs[j]
        excess += outer*xs[i]
    return excess*0.5

def redlich_kister_build_structure(N, shape, data, indexes):
    r'''Builds a redlich-kister compatible
    tensor (data structure) from pairs of indexes, and
    the coefficients associated with those indexes. 
    This is especially important because of the asymmetry
    of the model.

    Parameters
    ----------
    N : int
        The number of components in the mixture, [-]
    shape : tuple(int,) of dimensionality desired
        The parameter to specify the shape of each of the
        [i][j] component data structures, [-]
    data : list[float] or list[list[float]]
        The coefficients regressed or from literature, [various]
    indexes : list[tuple(int, int)]
        A list of index pairs. This is used to determine
        which components are associated with what data.
        Order matters! (i,j) != (j, i)

    Returns
    -------
    structure : list[list[data]] of shape [N, N, *shape]
        Output structurez

    Notes
    -----
    This is a fairly computationally intensive calculation
    and should be cached
    '''
    out = []
    
    if len(shape) == 1:
        one_d = True
        two_d = False
    elif len(shape) == 2:
        one_d = False
        two_d = True
    else:
        raise ValueError("Shape must be provided")
        
    if len(indexes) != len(data):
        raise ValueError("Index and data length must be the same")
        
    data_dict = {idx: d for idx, d in zip(indexes, data)}
        
    for i in range(N):
        l = []
        for j in range(N):
            if one_d:
                l.append([0.0]*shape[0])
            else:
                l.append([[0.0]*shape[1] for _ in range(shape[0])])
        out.append(l)
        
    for i in range(N):
        for j in range(N):
            if (i, j) in data_dict:
                thing = data_dict[(i, j)]
                out[i][j] = thing
            elif (j, i) in data_dict:
                thing = data_dict[(j, i)]
                if one_d:
                    thing = redlich_kister_reverse(thing)
                else:
                    thing = redlich_kister_reverse_2d(thing)
                out[i][j] = thing
    return out


def redlich_kister_T_dependence(structure, T, N, N_terms, N_T):
    r'''Compute a redlich-kister set of A coefficients from
    a temperature-dependent data struture with number
    of temperature-dependent terms `N_T`. Essentially
    this takes a 4d array and returns a 3d one.

    Parameters
    ----------
    structure : list[list[list[float]]] of shape [N, N, N_terms, N_T]
        Input structure with temperature dependence coefficients, [various]
    T : float
        Temperature, [K]
    N : int
        The number of components in the mixture, [-]
    N_terms : int
        The number of terms in the expansion, [-]
    N_T : int
        The number of terms in the temperature expansion, [-]

    Returns
    -------
    structure : list[list[data]] of shape [N, N, N_terms]
        Output structure

    Notes
    -----
    '''
    T2 = T*T
    Tinv = 1.0/T
    T2inv = Tinv*Tinv
    logT = log(T)
    out = [[[0.0]*N_terms for _ in range(N)] for _ in range(N)]

    if N_T == 2:
        for i in range(N):
            out_2d = out[i]
            in_3d = structure[i]
            for j in range(N):
                out_1d = out_2d[j]
                in_2d = in_3d[j]
                for k in range(N_terms):
                    in_1d = in_2d[k]
                    out_1d[k] = in_1d[0] + in_1d[1]*Tinv
    elif N_T == 3:
        for i in range(N):
            out_2d = out[i]
            in_3d = structure[i]
            for j in range(N):
                out_1d = out_2d[j]
                in_2d = in_3d[j]
                for k in range(N_terms):
                    in_1d = in_2d[k]
                    out_1d[k] = in_1d[0] + in_1d[1]*Tinv + in_1d[2]*logT
    elif N_T == 6:
        for i in range(N):
            out_2d = out[i]
            in_3d = structure[i]
            for j in range(N):
                out_1d = out_2d[j]
                in_2d = in_3d[j]
                for k in range(N_terms):
                    in_1d = in_2d[k]
                    out_1d[k] = in_1d[0] + in_1d[1]*Tinv + in_1d[2]*logT + in_1d[3]*T + in_1d[4]*T2inv + in_1d[5]*T2
    else:
        raise ValueError("Unsupported number of terms")
    return out


def redlich_kister_excess_inner_binary(ais, xs):
    r'''Compute the redlich-kister excess for a binary
    system. This calculation is optimized. This works
    with the same values of coefficients as `redlich_kister_excess_inner`
    but without the excess dimensionality of the input
    data. 

    Parameters
    ----------
    ais : list[float] of size N_terms
        `A` coefficients`, [-]
    xs : list[float]
        Binary mole fractions, [-]

    Returns
    -------
    excess : float
        The calculated excess value [-]

    Notes
    -----
    '''
    x0, x1 = xs
    x_diff = (x0 - x1)
    x_product = x0*x1
    terms = len(ais)
    GE = 0.0
    factor = 1.0
    for i in range(terms):
        GE += ais[i]*x_product*factor
        factor *= x_diff
    return GE

def redlich_kister_excess_binary(coefficients, x0, T, N_T, N_terms):
    T2 = T*T
    Tinv = 1.0/T
    T2inv = Tinv*Tinv
    logT = log(T)
    x1 = 1.0 - x0
    x_diff = (x0 - x1)
    x_product = x0*x1
    GE = 0.0
    factor = 1.0
    for i in range(N_terms):
        offset = i*N_T
        ai = coefficients[offset] 
        if N_T >= 2:
            ai +=  coefficients[offset+1]*Tinv
        if N_T >= 3:
            ai += coefficients[offset+2]*logT
        if N_T >= 4:
            ai +=  coefficients[offset+3]*T
        if N_T >= 5:
            ai += coefficients[offset+4]*T2inv
        if N_T >= 6:
            ai += coefficients[offset+5]*T2
        GE += ai*x_product*factor
        factor *= x_diff
    return GE

def redlich_kister_fitting_to_use(coeffs, N_terms, N_T):
    out = []
    for i in range(N_terms):
        l = []
        for j in range(N_T):
            l.append(coeffs[i*N_T + j])
        out.append(l)
    return out
